import itertools
import pprint
import datetime
import re
import logging

from copy import copy

log = logging.getLogger(__name__)
re_flags_separator = re.compile('[, ]+')

ATTR_PREFIX_FLAG = 'Flags'
ATTR_PREFIX_LINK = 'Link'
ATTR_PREFIX_NOTE = 'Note'


class Task(object):
    index = ''
    slug = ''
    name = ''
    note = ''
    priority = 500
    dStart = datetime.datetime.max
    dFinish = datetime.datetime.min
    dAcStart = datetime.datetime.max
    dAcFinish = datetime.datetime.min
    milestone = False
    p_complete = 0.0
    flags = []
    level = 1
    link = ''

    tasks = []

    resource = None

    _date_format = '%Y-%m-%dT%H:%M:%S'
    _rx = None
    _schedule = None
    _subtree_hash = None

    def __init__(self, schedule=None, level=1):
        self.tasks = []
        self.flags = []
        self._schedule = schedule
        self.p_complete = 0.0
        self.priority = 500
        self.milestone = False
        self.level = level

    def __unicode__(self):
        return '%s %s MS:%s  (%s - %s) ac(%s - %s)  F%s  [%s]' % (
            self.slug, self.name, self.milestone, self.dStart, self.dFinish,
            self.dAcStart, self.dAcFinish, self.flags, len(self.tasks))

    def __str__(self):
        return unicode(self).encode('utf-8')

    def parse_extended_attr(self, value, key=None):
        """
        According to given value it will guess if it is flag, link or note
        definition and fill proper task attribute.

        Args:
            value: string to parse, or set attribute specified by 'key' arg
            key: (optional) set value to task attribute according to this key
        """
        if not value:
            return

        if key:
            val = value
        else:
            pieces = value.split(':', 1)
            if len(pieces) != 2:
                # it's not a string in format 'Flag: qe, dev' - don't process
                return

            key, val = pieces
            key = key.strip().lower()
            val = val.strip()

        if key.lower() == ATTR_PREFIX_FLAG.lower():
            val = val.lower()
            flags = re_flags_separator.split(val)
            for flag in flags:
                flag = str(flag.strip())
                if flag:
                    self.flags.append(flag)
            self._schedule.used_flags |= set(self.flags)

        elif key.lower() == ATTR_PREFIX_LINK.lower():
            self.link = str(val)

        elif key.lower() == ATTR_PREFIX_NOTE.lower():
            # in case of multiple notes - concatenate
            self.note = ' '.join([self.note, val]).lstrip()

        else:
            log.warn('Extended attr "{}" wasn\'t recognized.'.format(key))

    @staticmethod
    def _workaround_it_phase_names(eTask):
        """Corrects phase names"""
        name_map = {'Concept': 'Concept Phase',
                    'Planning': 'Planning Phase',
                    'Develop': 'Development Phase',
                    'Test': 'Testing Phase',
                    'Launch': 'Launch Phase',
                    'Maintenance': 'Maintenance Phase',
                    }

        eName_list = eTask.xpath('Name')
        if eName_list:
            name = eName_list[0].text
            if name:
                name = name.strip()

            level = int(eTask.xpath('OutlineLevel')[0].text)

            if level in [1, 2] and name in name_map.keys():
                # check that there is no full named task else in the schedule
                root = eTask.getroottree()
                eFullnamed_task_list = root.xpath('//Task[Name = "%s"]' % name_map[name])

                if not eFullnamed_task_list:  # if not - map the name
                    return name_map[name]
            return name
        return ''

    def check_for_phase(self):
        PHASE_NAMES = ('planning phase', 'development phase', 'testing phase',
                       'launch phase', 'maintenance phase')
        if self.name.lower() in PHASE_NAMES:
            self.name = self.name.title()
            self._schedule.phases.append(self)

    @property
    def desc_tasks_count(self):
        count = len(self.tasks)
        for task in self.tasks:
            count += task.desc_tasks_count
        return count

    def get_type(self, check_same_start_finish=False):
        if self.tasks:
            return 'Container'
        elif self.milestone and (not check_same_start_finish or self.dAcStart == self.dAcFinish):
            return 'Milestone'
        else:
            return 'Task'

    def dump_as_dict(self, recursive=True):
        attrs = copy(vars(self))
        # avoid infinite looping schedule > task > schedule ...
        exclude = ['_schedule', '_subtree_hash']

        if recursive:
            attrs['tasks'] = []
            for task in self.tasks:
                attrs['tasks'].append(task.dump_as_dict())
        else:
            exclude.append('tasks')

        [attrs.pop(item) for item in exclude if item in attrs]

        return attrs

    @classmethod
    def load_from_dict(cls, input_dict, schedule):
        task = cls(schedule)

        for key, val in input_dict.items():
            task.__setattr__(key, val)

        task.tasks = []
        for new_task in input_dict.get('tasks', []):
            task.tasks.append(Task.load_from_dict(new_task, schedule))

        return task

    @property
    def subtree_hash(self):
        attrs = ['name', 'dStart', 'dFinish', 'dAcStart', 'dAcFinish']

        if self._subtree_hash is None:
            self._subtree_hash = ''

            for child_task in self.tasks:
                values_list = [str(getattr(child_task, attr)) for attr in attrs]
                self._subtree_hash += ''.join(values_list) + child_task.subtree_hash

        return self._subtree_hash


class Schedule(object):
    slug = ''
    name = ''  # Product 1.2
    tasks = []
    dStart = None
    dFinish = None
    changelog = {}
    mtime = None

    phases = []
    resources = {}
    assignments = []
    used_flags = None
    ext_attr = {}
    flags_attr_id = None
    id_reg = set()

    _eTask_list = []
    _task_index = 1
    _taskname_flat_registry = None

    errors_import = []

    def __init__(self):
        self.tasks = []
        self.phases = []
        self.used_flags = set()
        self.changelog = {}
        self.ext_attr = {}
        self.unique_id_re = re.compile('[\W_]+')
        self.errors_import = []
        self.mtime = None


    def make_flat(self):
        '''Convert tasks structure to flat list'''
        def add_tasks_to_list(tasks, tasks_list):
            for task in tasks:
                subtasks = copy(task.tasks)
                task.tasks = []
                tasks_list.append(task)
                add_tasks_to_list(subtasks, tasks_list)
        
        flat_tasks = []
        add_tasks_to_list(self.tasks, flat_tasks)
        self.tasks = flat_tasks


    def filter_milestones(self):
        '''Keep only tasks that are milestones or contain children that are'''
        def filter_tasks(tasks):
            tasks_list = []
            for task in tasks:
                filtered_subtasks = filter_tasks(task.tasks)
                
                logmsg_fmt = 'REMOVE: {} {}'
                
                if task.milestone or filtered_subtasks:
                    task.tasks = filtered_subtasks
                    tasks_list.append(task)
                    logmsg_fmt = 'ADD: {} {}'
                    
                log.info(logmsg_fmt.format(task.name, task.milestone))
                    
            return tasks_list
        
        self.tasks = filter_tasks(self.tasks)
        
    
    def filter_flags(self, show=None, hide=None):
        '''
        Keep only tasks that match flag conditions or contain children tha match
        
        Args:
            show: list of flags to show
            hide: list of flags to hide
        '''
        
        def filter_tasks(tasks):
            tasks_list = []
            for task in tasks:
                # see if there are any matching children
                filtered_subtasks = filter_tasks(task.tasks)

                # Add to result if not hidden and (has subtasks or fits show)
                logmsg_fmt = 'REMOVE: {} {}'
                if (not (set(task.flags) & set(hide)) 
                    and (filtered_subtasks 
                         or (set(task.flags) & set(show))
                         or not show)):
                    task.tasks = filtered_subtasks
                    tasks_list.append(task)
                    
                    logmsg_fmt = 'ADD: {} {}'
                    
                    self.used_flags.update(task.flags)
                
                log.info(logmsg_fmt.format(task.name, task.flags))
                
            
            return tasks_list
            
        if show is None:
            show = []
        if hide is None:
            hide = []
            
        if show or hide:
            self.used_flags = set()
            
            log.info('FLAG filter: SHOW {}   HIDE {}'.format(show, hide))

            self.tasks = filter_tasks(self.tasks)


    def check_top_task(self):
        def _raise_index(task):
            task.index = int(task.index) + 1
            for t in task.tasks:
                _raise_index(t)

        if len(self.tasks) <= 1:
            return

        # need one top task
        top_task = Task(self)
        top_task.index = 1
        #top_task.name = '%s %s' % (self.name, self.version)
        # removed version as per BZ#1396303 - see how that works
        top_task.name = self.name
        top_task.slug = self.slug

        for task in self.tasks:
            if top_task.dStart:
                top_task.dStart = min(top_task.dStart,
                                      task.dStart,
                                      task.dAcStart)
            else:
                top_task.dStart = task.dStart

            top_task.dAcStart = top_task.dStart

            if top_task.dFinish:
                top_task.dFinish = max(top_task.dFinish,
                                       task.dAcFinish,
                                       task.dFinish)
            else:
                top_task.dFinish = task.dFinish

            top_task.dAcFinish = top_task.dFinish

            _raise_index(task)
            top_task.tasks.append(task)

        self.tasks = [top_task]
        
    
    def generate_slugs(self):
        def gen_slugs_recurse(tasks, id_prefix):
            for task in tasks:
                task.slug = self.get_unique_id(task.name, id_prefix)
                gen_slugs_recurse(task.tasks, task.slug)
        
        self.id_reg = set()
        gen_slugs_recurse(self.tasks, self.slug)
    

    def print_tasks(self, tasks=None, level=0):
        if tasks is None:
            tasks = self.tasks
        for task in tasks:
            print level * '  ', str(task)
            if task.tasks:
                self.print_tasks(task.tasks, level + 1)

    @staticmethod
    def _diff_task_attrs(left, right, attrs, whole_days=False):
        diff = {}

        for a in attrs:
            left_val = left.__getattribute__(a)
            right_val = right.__getattribute__(a)
            if left_val != right_val:
                try:
                    if whole_days and \
                                    left_val.day == right_val.day and \
                                    left_val.month == right_val.month and \
                                    left_val.year == right_val.year:
                        # When we compare just whole days, we don't mind
                        # to have diff in time
                        continue
                except AttributeError:
                    pass

                diff[a] = (str(left_val), str(right_val))
        return diff

    def _diff_tasks(self, left, right, attrs=None, whole_days=False):
        default_attrs = ['name', 'dStart', 'dFinish', 'dAcStart', 'dAcFinish',
                         'link', 'note']
        if not attrs:
            attrs = default_attrs
        ret = ''
        missing_str = ('{}: Some phase upwards is missing to compare with '
                       'this one on the {} side.\n')

        for left, right in itertools.izip_longest(left, right):
            if not left:
                ret += missing_str.format(right.name, 'LEFT')
                break
            if not right:
                ret += missing_str.format(left.name, 'RIGHT')
                break

            d = self._diff_task_attrs(left, right, attrs, whole_days)
            if len(d):
                name_warning = ''
                if 'name' in d.keys():
                    name_warning = (' <--- Name of the phase doesn\'t match. '
                                    'Missing phase?')
                # width=1 force wrap tuple to newline
                ret += '{}:{}\n{}\n'.format(
                    left.name,
                    name_warning,
                    pprint.pformat(d, width=1)
                )

            ret += self._diff_tasks(left.tasks, right.tasks, attrs, whole_days)
        return ret

    def _diff_schedule_attrs(self, schedule, whole_days=False):
        attrs = ['name', 'changelog', 'dStart', 'dFinish', 'assignments', 'used_flags']
        ret = ''
        diff = dict()

        for attr in attrs:
            self_s = self.__getattribute__(attr)
            schedule_s = schedule.__getattribute__(attr)

            if self_s != schedule_s:
                try:
                    if whole_days and \
                                    self_s.day == schedule_s.day and \
                                    self_s.month == schedule_s.month and \
                                    self_s.year == schedule_s.year:
                        # When we compare just whole days, we don't mind
                        # to have diff in time
                        continue
                except AttributeError:
                    pass
                diff[attr] = (str(self_s), str(schedule_s))

        if len(diff):
            # width=1 force wrap tuple to newline
            ret += 'Schedule attrs:\n'
            ret += pprint.pformat(diff, width=1)
        return ret

    def _diff_schedule_version(self, schedule):
        attrs = self._version.keys()
        ret = ''
        diff = dict()

        for attr in attrs:
            self_s = self._version[attr]
            schedule_s = self._version[attr]

            if self_s != schedule_s:
                diff[attr] = (self_s, schedule_s)

        if len(diff):
            # width=1 force wrap tuple to newline
            ret += 'Schedule version:\n'
            ret += pprint.pformat(diff, width=1)

        return ret

    def diff(self, schedule, attrs=None, whole_days=False):
        """
        Really simple diff to match two schedules.

        @param schedule Instance to another schedule to figure out differences
        @param attrs Task attributes to check
        @param whole_days Flag to ignore time differences of dates
        """
        ret = ''
        diff_tasks = self._diff_tasks(self.tasks, schedule.tasks, attrs,
                                      whole_days)
        if diff_tasks:
            ret += 'Tasks attrs:\n'
            ret += diff_tasks
        ret += self._diff_schedule_attrs(schedule, whole_days=whole_days)

        return ret.strip()

    def slugify_str(self, orig_str):
        return self.unique_id_re.sub('_', orig_str.lower())


    def get_unique_id(self, orig_str, id_prefix=''):
        '''Return unique id within schedule'''
        
        # shortcut - first orig_str gets prefix        
        if not self.id_reg and id_prefix:
            self.id_reg.add(id_prefix)
            return id_prefix

        pref = copy(id_prefix)
        pref += '.'

        source = self.slugify_str(orig_str)

        found = False
        test_id = ''
        for word in source.split('_'):
            test_id = pref + word
            if test_id in self.id_reg:
                pref = test_id + '_'
            else:
                found = True
                self.id_reg.add(test_id)
                break

        if not found:
            # duplicate orig_str names - add numbering
            n = 2
            while '%s_%s' % (test_id, n) in self.id_reg:
                n += 1

            test_id = '%s_%s' % (test_id, n)
            self.id_reg.add(test_id)

            log.info('Duplicate Names: %s, adding: %s' % (source, test_id))

        return test_id

    def dump_as_dict(self):
        schedule = copy(vars(self))

        schedule['tasks'] = []
        for task in self.tasks:
            schedule['tasks'].append(task.dump_as_dict())

        schedule['phases'] = []
        for phase in self.phases:
            schedule['phases'].append(phase.dump_as_dict(recursive=False))

        # set() is not serializable into json
        schedule['used_flags'] = sorted(list(self.used_flags))
        return schedule

    @classmethod
    def load_from_dict(cls, input_dict):
        # This will preserve reference to original class attributes
        schedule = cls()
        for key, val in input_dict.items():
            schedule.__setattr__(key, val)

        schedule.tasks = []
        for task in input_dict['tasks']:
            schedule.tasks.append(Task.load_from_dict(task, schedule))

        schedule.phases = []
        for phase in input_dict['phases']:
            schedule.phases.append(Task.load_from_dict(phase, schedule))

        schedule.used_flags = set(input_dict['used_flags'])

        return schedule

    def _build_tasks_flat_registry(self, tasks_param):
        if self._taskname_flat_registry is None:
            self._taskname_flat_registry = set()

        for task in tasks_param:
            self._taskname_flat_registry.add(task.name)
            self._build_tasks_flat_registry(task.tasks)

    def check_for_taskname(self, tasks):
        """
        Method consume list of task names to check their existence specified
        as dict
        Args:
            tasks: dict specification of task names to check. Example:
            {'exactTaskname': False, 'matchBeginningTaskname': True}

        Returns:
            Set of task names, that haven't been found in schedule.

        """
        if not isinstance(tasks, dict):
            raise Exception(
                "'tasks' argument has to be dict, for example: {'task1': False,"
                " 'matchBeginningTaskname': True}")

        if not self._taskname_flat_registry:
            self._build_tasks_flat_registry(self.tasks)

        missing_tasks = set()
        for task_name, start_with_flag in tasks.items():
            if start_with_flag:
                re_task_str = re.escape(task_name)
                re_task = re.compile(re_task_str + '.*')
                found = False

                for task in self._taskname_flat_registry:
                    if re_task.match(task):
                        found = True
                        break
                if not found:
                    missing_tasks.add(task_name)
            elif task_name not in self._taskname_flat_registry:
                missing_tasks.add(task_name)

        return missing_tasks



from schedules_tools import converter
from schedules_tools.models import Schedule
from schedules_tools.storage_handlers import StorageBase
#from scripts import schedule_converter as converter_cli
import tempfile
import os
import datetime
import mock
import pytest

DATA_DIR = 'schedule_files'

PARENT_DIRNAME = os.path.basename(os.path.dirname(os.path.realpath(__file__)))
BASE_DIR = os.path.dirname(os.path.realpath(
    os.path.join(__file__, os.pardir)))
CURR_DIR = os.path.join(BASE_DIR, PARENT_DIRNAME)


class BaseTestConvert(object):
    _filenames = {
        'tjx': 'proj-10-1-2.tjx',
        'tjx2': 'proj-10-1-2-v2.tjx',
        'smartsheet': 'import-schedule-msp.xml'
    }
    file_tjx = ''
    file_tjx2 = ''
    file_smartsheet = ''
    schedule_name = 'Test project 10'

    file_out_fd = None
    file_out_name = None

    @pytest.fixture(autouse=True)
    def setUp(self):
        self.file_out_fd, self.file_out_name = tempfile.mkstemp()

        # prepend data dir to real path of files
        for k, v in self._filenames.items():
            key = 'file_{}'.format(k)
            val = os.path.join(CURR_DIR, DATA_DIR, v)
            val = os.path.realpath(val)
            setattr(self, key, val)
        yield
        os.remove(self.file_out_name)


class TestConverter(BaseTestConvert):

    def test_init_storage_handler(self):
        handle = 'source.tjx'

        conv = converter.ScheduleConverter()
        assert conv.storage_handler is None

        conv._init_storage_handler(handle, storage_src_format='cvs')
        inst_storage = conv.storage_handler

        assert isinstance(inst_storage, StorageBase)

        # try to reinitialize
        conv._init_storage_handler(handle, storage_src_format='cvs')
        assert conv.storage_handler is inst_storage

    @mock.patch('schedules_tools.converter.ScheduleConverter.storage_handler',
                new_callable=mock.PropertyMock)
    @mock.patch.object(converter.ScheduleConverter, '_get_schedule_handler_cls')
    @mock.patch.object(converter.ScheduleConverter, '_get_local_handle_from_storage')
    def test_handle_modified_since_from_handle(self,
                                               mock_get_local_handle_from_storage,
                                               mock_get_schedule_handler_cls,
                                               mock_storage_handler):
        conv = converter.ScheduleConverter()
        handle = 'source.tjx'
        curr_date = datetime.datetime(2010, 3, 13)
        mock_storage_handler.return_value = None

        conv.handle_modified_since(handle, curr_date)

        mock_get_schedule_handler_cls.assert_called()
        mock_storage_handler.get_handle_mtime.assert_not_called()

    @mock.patch.object(converter.ScheduleConverter, 'storage_handler')
    @mock.patch.object(converter.ScheduleConverter, '_get_schedule_handler_cls')
    @mock.patch.object(converter.ScheduleConverter, '_get_local_handle_from_storage')
    def test_handle_modified_since_from_storage(self,
                                                mock_get_local_handle_from_storage,
                                                mock_get_schedule_handler_cls,
                                                mock_storage_handler):
        handle = 'source.tjx'
        date_storage_handler = datetime.datetime(2010, 3, 13)

        mock_provide_mtime = mock.PropertyMock()
        mock_provide_mtime.return_value = True
        mock_storage_handler.get_handle_mtime.return_value = date_storage_handler
        type(mock_storage_handler).provide_mtime = mock_provide_mtime

        conv = converter.ScheduleConverter()
        conv.handle_modified_since(handle, date_storage_handler)
        mock_storage_handler.handle_modified_since.assert_called()

    @mock.patch.object(converter.ScheduleConverter, 'storage_handler')
    @mock.patch.object(converter.ScheduleConverter, '_get_schedule_handler_cls')
    @mock.patch.object(converter.ScheduleConverter, '_get_local_handle_from_storage')
    def test_import_schedule_with_storage_provides_mtime_changelog(
            self,
            mock_get_local_handle_from_storage,
            mock_get_schedule_handler_cls,
            mock_storage_handler):
        mock_schedule_handler_cls = mock.Mock()
        mock_schedule_handler = mock.Mock()
        provided_schedule = Schedule()
        date_mtime_storage = datetime.datetime(2009, 3, 13)
        provided_schedule.mtime = datetime.datetime(2019, 3, 13)

        mock_schedule_handler.import_schedule.return_value = provided_schedule
        type(mock_schedule_handler).provide_changelog = mock.PropertyMock(return_value=False)

        mock_schedule_handler_cls.return_value = mock_schedule_handler
        mock_get_schedule_handler_cls.return_value = mock_schedule_handler_cls

        mock_storage_handler.get_handle_mtime.return_value = date_mtime_storage
        type(mock_storage_handler).provide_mtime = mock.PropertyMock(return_value=True)

        mock_storage_handler.get_handle_changelog.return_value = 'storage-changelog'
        type(mock_storage_handler).provide_changelog = mock.PropertyMock(return_value=True)

        handle = 'source.tjx'

        conv = converter.ScheduleConverter()
        schedule = conv.import_schedule(handle)

        assert schedule is provided_schedule
        assert schedule.mtime == date_mtime_storage
        assert schedule.changelog == 'storage-changelog'
        mock_schedule_handler.get_handle_changelog.assert_not_called()
        mock_storage_handler.get_handle_changelog.assert_called()
        mock_storage_handler.get_handle_mtime.assert_called()

    @mock.patch.object(converter.ScheduleConverter, 'storage_handler')
    @mock.patch.object(converter.ScheduleConverter, '_get_schedule_handler_cls')
    @mock.patch.object(converter.ScheduleConverter, '_get_local_handle_from_storage')
    def test_import_schedule_schedule_handler_provides_changelog(
            self,
            mock_get_local_handle_from_storage,
            mock_get_schedule_handler_cls,
            mock_storage_handler):
        mock_schedule_handler_cls = mock.Mock()
        mock_schedule_handler = mock.Mock()
        provided_schedule = Schedule()
        date_mtime_storage = datetime.datetime(2009, 3, 13)
        provided_schedule.mtime = datetime.datetime(2019, 3, 13)

        mock_schedule_handler.import_schedule.return_value = provided_schedule
        type(mock_schedule_handler).provide_changelog = mock.PropertyMock(return_value=True)
        mock_schedule_handler.get_handle_changelog.return_value = 'sch-changelog'

        mock_schedule_handler_cls.return_value = mock_schedule_handler
        mock_get_schedule_handler_cls.return_value = mock_schedule_handler_cls

        mock_storage_handler.get_handle_mtime.return_value = date_mtime_storage
        type(mock_storage_handler).provide_mtime = mock.PropertyMock(return_value=True)

        mock_storage_handler.get_handle_changelog.return_value = 'storage-changelog'
        type(mock_storage_handler).provide_changelog = mock.PropertyMock(return_value=True)

        handle = 'source.tjx'

        conv = converter.ScheduleConverter()
        schedule = conv.import_schedule(handle)

        assert schedule is provided_schedule
        assert schedule.mtime == date_mtime_storage
        assert schedule.changelog == 'storage-changelog'
        mock_storage_handler.get_handle_changelog.assert_called()
        mock_storage_handler.get_handle_mtime.assert_called()

    @mock.patch.object(converter.ScheduleConverter, 'storage_handler')
    @mock.patch.object(converter.ScheduleConverter, '_get_schedule_handler_cls')
    @mock.patch.object(converter.ScheduleConverter, '_get_local_handle_from_storage')
    def test_import_schedule_storage_handler_overrides_schedule_changelog(
            self,
            mock_get_local_handle_from_storage,
            mock_get_schedule_handler_cls,
            mock_storage_handler):
        mock_schedule_handler_cls = mock.Mock()
        mock_schedule_handler = mock.Mock()
        provided_schedule = Schedule()
        date_mtime_storage = datetime.datetime(2009, 3, 13)
        provided_schedule.mtime = datetime.datetime(2010, 3, 13)

        mock_schedule_handler.import_schedule.return_value = provided_schedule
        type(mock_schedule_handler).provide_changelog = mock.PropertyMock(return_value=True)
        mock_schedule_handler.get_handle_changelog.return_value = 'sch-changelog'

        mock_schedule_handler_cls.return_value = mock_schedule_handler
        mock_get_schedule_handler_cls.return_value = mock_schedule_handler_cls

        type(mock_storage_handler).provide_mtime = mock.PropertyMock(return_value=False)

        mock_storage_handler.get_handle_changelog.return_value = 'storage-changelog'
        type(mock_storage_handler).provide_changelog = mock.PropertyMock(return_value=True)

        handle = 'source.tjx'
        conv = converter.ScheduleConverter()
        schedule = conv.import_schedule(handle)

        assert schedule is provided_schedule
        assert schedule.changelog == 'storage-changelog'
        mock_storage_handler.get_handle_changelog.assert_called()
        mock_storage_handler.get_handle_mtime.assert_not_called()

    @mock.patch.object(converter.ScheduleConverter, 'storage_handler')
    def test_cleanup_local_handle(
            self,
            mock_storage_handle):
        conv = converter.ScheduleConverter()
        conv.cleanup_local_handle()

        mock_storage_handle.clean_local_handle.assert_called()

    @mock.patch('schedules_tools.converter.ScheduleConverter.storage_handler',
                new_callable=mock.PropertyMock)
    def test_cleanup_local_handle_without_storage(
            self,
            mock_storage_handle):
        mock_storage_handle.return_value = None
        conv = converter.ScheduleConverter()
        conv.cleanup_local_handle()

        mock_storage_handle.clean_local_handle.assert_not_called()

    def test_msp_parse_flags_note_url(self):
        conv = converter.ScheduleConverter()
        schedule = conv.import_schedule(self.file_smartsheet)

        # Test1 - Testing Phase task
        assert ['flag1'] == schedule.tasks[0].tasks[0].tasks[2].flags

        # Test2 - Another task
        assert ['flag1', 'flag2', 'flag3'] == schedule.tasks[0].tasks[1].tasks[1].flags

        # Test2 - Another task
        assert 'test2 note' == schedule.tasks[0].tasks[1].tasks[1].note

        # Test1 - Development
        assert 'https://github.com/1' == schedule.tasks[0].tasks[0].tasks[0].link


#class TestConverterCLI(BaseTestConvert):
#    def test_discover_handlers(self):
#        args = ['--handlers-path', 'tests/foohandlers',
#                self.file_tjx, 'abc', self.file_out_name]
#        converter_cli.main(args)
#        with open(self.file_out_name) as fd:
#            line = fd.readline()
#            assert line == self.schedule_name
#
#    def test_override_handlers(self):
#        args = ['--handlers-path', 'tests/conflicthandlers',
#                self.file_tjx2, 'tjx', self.file_out_name]
#        converter_cli.main(args)
#        with open(self.file_out_name) as fd:
#            line = fd.readline()
#            assert line == 'schedule.name={}'.format(self.schedule_name)
#
#    def test_tjx_msp(self):
#        args = [self.file_tjx, 'msp', self.file_out_name]
#        converter_cli.main(args)



'''
Created on Feb 15, 2012

@author: pslama
'''
import datetime

from subprocess import Popen, PIPE
from setup_utils import get_rpm_version, get_rpm_version_format, get_git_tag_list


CHANGELOG_FILENAME_PATH = 'spec/changelog.txt'

def get_last_changelog_commit(filename):
    """
    Get last string sequence after dot character from the first line of the filename.
    Returns None if the file does not exist.
    """
    try:
        fd = open(filename, 'r')
    except IOError:
        return None

    first_line = fd.readline().strip()
    fd.close()
    last_commit = first_line.rsplit('.')[-1]
    return last_commit


def get_rpm_log():
    # get existing tags reversed
    try:
        first_tag_commit = get_git_tag_list()[-1]
    except IndexError:
        # no tag - version 1.0.0
        print 'NO git tag where to take version from! Pls set tag.'
        exit()
    
    last_commit =  get_last_changelog_commit(CHANGELOG_FILENAME_PATH)
    if last_commit:
        first_tag_commit = last_commit
    
    proc = Popen(['git', 'log', "--branches='master or stable'",
                  "--pretty=format:* %ct %an <%ae> %h%n- %s%n",
                  '%s..' % first_tag_commit],
                 stdout=PIPE, stderr=PIPE)
    output, err = proc.communicate()
    ver_lines = output.splitlines()
    out = {}
    
    for ver_line in ver_lines:
        if not len(ver_line):
            continue
        elif ver_line[0] == '*':
            timestamp, sep, rest = ver_line[2:].partition(' ')
            actual_ts = timestamp
            rest, sep, commit_hash = rest.rpartition(' ')
            #TODO:3 git log -3 --pretty --walk-reflogs master stable
            date = datetime.date.fromtimestamp(float(timestamp))
            beg_date = date.strftime('%a %b %d %Y')
            ver_date = date.strftime('%Y%m%d')
            new_line = '* %s %s %s' % (
                    beg_date, rest, get_rpm_version_format(get_rpm_version(commit_hash)))
            out[actual_ts] = new_line + '\n'
        elif actual_ts:
            out[actual_ts] += ver_line + '\n'
    
    output = [out[key] for key in sorted(out.iterkeys(), reverse=True)]
    output.append('')
    output = "\n".join(output)
    
    if last_commit:
        fd = open(CHANGELOG_FILENAME_PATH, 'r')
        old_changelog = fd.read()
        fd.close()
        output = "{0}{1}".format(output if output else '', old_changelog)
    
        fd = open(CHANGELOG_FILENAME_PATH, 'w')
        fd.write(output)
        fd.close()
    else:
        fd = open(CHANGELOG_FILENAME_PATH, 'w')
        fd.write(output)
        fd.close()
    
    return output

if __name__ == '__main__':
    print(get_rpm_log())
    



import datetime
import mock
import os
import pytest

from schedules_tools.storage_handlers import cvs as cvs_mod
from schedules_tools import storage_handlers
from schedules_tools.tests import jsondate

BASE_DIR = os.path.dirname(os.path.realpath(__file__))


class BaseCvsTest(object):
    handler_class = cvs_mod

    @classmethod
    def _make_reference_obj(cls, handle=None, checkout_dir=None, options=dict()):
        options = {
            'cvs_repo_name': 'repo',
            'cvs_root': ':gserver:someuser@cvs.myserver.com:/cvs/repo',
            'cvs_checkout_path': checkout_dir,
        }
        reference = cls.handler_class.StorageHandler_cvs(handle, options=options)
        return reference


class BaseCvsWithDefaultHandler(BaseCvsTest):
    reference_obj = None
    handle = 'program/rhel/rhel-7-0-0/rhel-7-0-0.tjp'
    checkout_dir = '/tmp/mycheckoutdir'

    @pytest.fixture(autouse=True)
    def setUp(self):
        self.reference_obj = self._make_reference_obj(self.handle,
                                                      self.checkout_dir)


class TestCvsCustomRepo(BaseCvsTest):
    def test_get_local_shared_path(self):
        checkout_dir = '/tmp/aaa'
        reference = self._make_reference_obj(checkout_dir=checkout_dir)
        path = 'program/release'

        ret = reference.get_local_shared_path(path)

        assert ret == '/tmp/aaa/repo/program/release'

    @mock.patch('schedules_tools.storage_handlers.cvs.os')
    def test_is_valid_cvs_dir_root_negative(self, mock_os):
        options = {
            'cvs_repo_name': 'program',
            'cvs_root': ':gserver:someuser@cvs.myserver.com:/cvs/reporoot'
        }
        path = 'program/product/release-1-2-3'

        reference = cvs_mod.StorageHandler_cvs(options=options)

        mock_fileopen = mock.mock_open(read_data=path)
        mock_cvs_repo = mock_fileopen
        cvs_root = ':gserver:otheruser@cvs.myserver.com:/cvs/reporoot/abc/def/h'
        mock_cvs_root = mock.mock_open(read_data=cvs_root)
        mock_fileopen.side_effect = [
            mock_cvs_repo.return_value,
            mock_cvs_root.return_value]

        mock_os.path.isdir.return_value = True
        with mock.patch('__builtin__.open', mock_fileopen):
            assert not reference._is_valid_cvs_dir(path)

    @mock.patch.object(cvs_mod.StorageHandler_cvs, 'refresh_local')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, 'get_local_shared_path')
    @mock.patch('schedules_tools.storage_handlers.cvs.os.path.getmtime')
    @mock.patch('schedules_tools.storage_handlers.cvs.os')
    def test_get_handle_mtime(self,
                              mock_os,
                              mock_getmtime,
                              mock_get_local_shared_path,
                              mock_refresh_local):
        handle = 'program/product/release/release.tjp'

        # 1477535432.123 == datetime.datetime(2016, 10, 27, 4, 30, 32, 123000)
        referece_ret = 1477535432.123
        mock_getmtime.return_value = referece_ret

        reference = self._make_reference_obj(handle=handle)

        ret = reference.get_handle_mtime()

        mock_refresh_local.assert_called()
        assert ret == datetime.datetime(2016, 10, 27, 4, 30, 32)
        mock_get_local_shared_path.assert_called_with(handle)

    def test_process_path(self):
        handle = 'program/fedora/f-25/f-25.xml'
        reference = self._make_reference_obj(handle=handle)
        assert reference._process_path() == 'program/fedora/f-25'


class TestCvs(BaseCvsWithDefaultHandler):
    @mock.patch('schedules_tools.storage_handlers.cvs.subprocess.Popen')
    def test_cvs_command(self, mock_popen):
        cmd = 'update program/rhel'

        mock_communicate = mock.Mock()
        mock_communicate.communicate.return_value = ('std-out', 'std-err')
        mock_popen.return_value = mock_communicate
        type(mock_communicate).returncode = mock.PropertyMock(return_value=0)

        return_value = self.reference_obj._cvs_command(cmd)
        assert return_value == ('std-out', 'std-err')
        mock_popen.assert_called()

    @mock.patch('schedules_tools.storage_handlers.cvs.subprocess.Popen')
    def test_cvs_command_nonzero_exit_code(self, mock_popen):
        cmd = 'update program/rhel'

        mock_communicate = mock.Mock()
        mock_communicate.communicate.return_value = ('std-out', 'std-err')
        mock_popen.return_value = mock_communicate
        type(mock_communicate).returncode = mock.PropertyMock(return_value=1)

        with pytest.raises(cvs_mod.CvsCommandException):
            self.reference_obj._cvs_command(cmd)
        mock_popen.assert_called()

    @mock.patch('schedules_tools.storage_handlers.cvs.remove_tree')
    @mock.patch('schedules_tools.storage_handlers.cvs.os')
    def test_wipeout_dir(self, mock_os, mock_remove_tree):
        path = 'somefile'
        mock_os.path.exists.return_value = True

        cvs_mod.StorageHandler_cvs._wipe_out_dir(path)

        mock_os.path.exists.assert_called_with(path)
        mock_remove_tree.assert_called_with(path)

    @mock.patch('schedules_tools.storage_handlers.cvs.remove_tree')
    @mock.patch('schedules_tools.storage_handlers.cvs.os')
    def test_wipeout_dir_not_exists(self, mock_os, mock_remove_tree):
        path = 'somefile'
        mock_os.path.exists.return_value = False

        cvs_mod.StorageHandler_cvs._wipe_out_dir(path)

        mock_os.path.exists.assert_called_with(path)
        mock_remove_tree.assert_not_called()

    @mock.patch('schedules_tools.storage_handlers.cvs.os')
    def test_is_valid_cvs_dir(self, mock_os):
        path = 'repo/product/release-1-2-3'

        mock_fileopen = mock.mock_open(read_data=path)
        mock_cvs_repo = mock_fileopen
        cvs_root = ':gsrv:test@cvs.myserver.com:/cvs/repo'
        mock_cvs_root = mock.mock_open(read_data=cvs_root)
        mock_fileopen.side_effect = [
            mock_cvs_repo.return_value,
            mock_cvs_root.return_value]

        mock_os.path.exists.return_value = True
        mock_os.path.isdir.return_value = True
        with mock.patch('__builtin__.open', mock_fileopen):
            assert self.reference_obj._is_valid_cvs_dir(path)

    @mock.patch('schedules_tools.storage_handlers.cvs.os.path.exists')
    @mock.patch('schedules_tools.storage_handlers.cvs.os.path.isdir')
    def test_is_valid_cvs_dir_CVSdir_negative(self, mock_isdir, mock_exists):
        path = 'program/aa/bbb'
        mock_isdir.return_value = False
        mock_exists.return_value = True

        assert not self.reference_obj._is_valid_cvs_dir(path)
        mock_isdir.assert_called_with('program/aa/bbb/CVS')

    @mock.patch('schedules_tools.storage_handlers.cvs.os.path.exists')
    def test_is_valid_cvs_dir_not_exists_negative(self, mock_exists):
        path = 'program/aa/bbb'
        mock_exists.return_value = False

        assert not self.reference_obj._is_valid_cvs_dir(path)
        mock_exists.assert_called_with('program/aa/bbb')

    @mock.patch('schedules_tools.storage_handlers.cvs.os.path.isdir')
    def test_is_valid_cvs_dir_repository_negative(self, mock_isdir):
        path = 'program/aa/bbb'
        mock_isdir.return_value = True
        mock_fileopen = mock.mock_open(read_data='asdf')

        with mock.patch('__builtin__.open', mock_fileopen):
            assert not self.reference_obj._is_valid_cvs_dir(path)

    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_cvs_checkout')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_update_shared_repo')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_is_valid_cvs_dir')
    @mock.patch('schedules_tools.storage_handlers.cvs.os')
    def test_refresh_local(self,
                           mock_os,
                           mock__is_valid_cvs_dir,
                           mock__update_shared_repo,
                           mock__cvs_checkout):
        mock_os.path.exists.return_value = True
        mock__is_valid_cvs_dir.return_value = True

        self.reference_obj.refresh_local()

        mock__update_shared_repo.assert_called()
        mock__cvs_checkout.assert_not_called()

    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_cvs_checkout')
    @mock.patch('schedules_tools.storage_handlers.cvs.os')
    def test_refresh_local_missing_shared_dir(
            self,
            mock_os,
            mock__cvs_checkout):
        mock_os.path.exists.return_value = False

        self.reference_obj.refresh_local()

        mock__cvs_checkout.assert_called()

    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_cvs_checkout')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_is_valid_cvs_dir')
    @mock.patch('schedules_tools.storage_handlers.cvs.os')
    def test_refresh_local_invalid_shared_dir(
            self,
            mock_os,
            mock__is_valid_cvs_dir,
            mock__cvs_checkout):
        mock_os.path.exists.return_value = True
        mock__is_valid_cvs_dir.return_value = False

        self.reference_obj.refresh_local()

        mock__cvs_checkout.assert_called()

    @mock.patch('tempfile.mkdtemp')
    @mock.patch('schedules_tools.storage_handlers.cvs.copy_tree')
    def test_copy_subtree_to_tmp(self, mock_copy_tree, mock_mkdtemp):
        process_path = 'rhel'
        mock_mkdtemp.return_value = '/tmp/asdf'

        return_value = self.reference_obj._copy_subtree_to_tmp(process_path)

        assert return_value == '/tmp/asdf'
        mock_copy_tree.assert_called_with('/tmp/mycheckoutdir/repo/rhel',
                                          '/tmp/asdf/rhel')

    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_copy_subtree_to_tmp')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, 'refresh_local')
    def test_get_local_handle(self,
                              mock_refresh_local,
                              mock_copy_subtree_to_tmp):
        mock_copy_subtree_to_tmp.return_value = '/tmp/tmpaaa'

        return_value = self.reference_obj.get_local_handle()
        return_ref = '/tmp/tmpaaa/program/rhel/rhel-7-0-0/rhel-7-0-0.tjp'
        assert return_value == return_ref
        assert self.reference_obj.tmp_root == '/tmp/tmpaaa'
        mock_refresh_local.assert_called()

    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_copy_subtree_to_tmp')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, 'refresh_local')
    def test_get_local_handle_multiple_times(self,
                                             mock_refresh_local,
                                             mock_copy_subtree_to_tmp):
        #mock_copy_subtree_to_tmp.return_value = '/tmp/tmpaaa'
        mock_copy_subtree_to_tmp.side_effect = ['/tmp/tmpaaa', '/tmp/tmpbbb', '/tmp/tmpccc']

        return_value = self.reference_obj.get_local_handle()
        return_ref = '/tmp/tmpaaa/program/rhel/rhel-7-0-0/rhel-7-0-0.tjp'
        assert return_value == return_ref
        assert self.reference_obj.tmp_root == '/tmp/tmpaaa'
        mock_refresh_local.assert_called()

        return_value = self.reference_obj.get_local_handle()
        assert return_value == return_ref
        assert self.reference_obj.tmp_root == '/tmp/tmpaaa'
        assert mock_refresh_local.call_count == 1

    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_copy_subtree_to_tmp')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, 'refresh_local')
    @mock.patch('schedules_tools.storage_handlers.cvs.remove_tree')
    def test_clean_local_handle(self,
                                mock_remove_tree,
                                mock_refresh_local,
                                mock_copy_subtree_to_tmp):
        self.reference_obj.get_local_handle()
        self.reference_obj.clean_local_handle()
        mock_remove_tree.assert_called()

        self.reference_obj.clean_local_handle()
        self.reference_obj.clean_local_handle()
        self.reference_obj.clean_local_handle()
        self.reference_obj.clean_local_handle()
        assert mock_remove_tree.call_count == 1

    def test_get_local_handle_get_specific_revision_by_rev(self):
        rev_number = '1.0'
        with pytest.raises(cvs_mod.NotImplementedFeature):
            self.reference_obj.get_local_handle(revision=rev_number)

    def test_get_local_handle_get_specific_revision_by_date(self):
        rev_date = datetime.datetime(2020, 3, 25)
        with pytest.raises(cvs_mod.NotImplementedFeature):
            self.reference_obj.get_local_handle(datetime=rev_date)

    def test_get_local_handle_get_specific_revision_by_rev_and_date(self):
        rev_number = '1.0'
        rev_date = datetime.datetime(2020, 3, 25)
        with pytest.raises(cvs_mod.NotImplementedFeature):
            self.reference_obj.get_local_handle(revision=rev_number,
                                                datetime=rev_date)

    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_cvs_command')
    def test_cvs_update(self, mock_cvs_command):
        mock_cvs_command.return_value = ('std-out', 'std-err')

        assert self.reference_obj._cvs_update() == ('std-out', 'std-err')
        mock_cvs_command.assert_called_with('update -dP', exclusive=True)

    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_cvs_command')
    def test_cvs_update_revdate(self, mock_cvs_command):
        mock_cvs_command.return_value = ('std-out', 'std-err')

        revision = datetime.date(2011, 9, 25)
        filename = 'myfile'
        output = self.reference_obj._cvs_update(filename, datetime_rev=revision)
        assert output == ('std-out', 'std-err')
        mock_cvs_command.assert_called_with(
            'update -dP -D "2011-09-25 00:00" myfile', exclusive=True)

    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_cvs_command')
    def test_cvs_update_revnumber(self, mock_cvs_command):
        mock_cvs_command.return_value = ('std-out', 'std-err')

        revision = '1.27'
        filename = 'myfile'
        output = self.reference_obj._cvs_update(filename, revision=revision)
        assert output == ('std-out', 'std-err')
        mock_cvs_command.assert_called_with(
            'update -dP -r "1.27" myfile', exclusive=True)

    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_cvs_command')
    def test_cvs_update_revnumber_revdate(self, mock_cvs_command):
        mock_process = mock.Mock()
        mock_process.communicate.return_value = ('std-out', 'std-err')
        type(mock_process).returncode = mock.PropertyMock(return_value=0)
        mock_cvs_command.return_value = mock_process

        revision = '1.2.3'
        revision_date = datetime.date(2011, 9, 25)
        filename = 'myfile'
        with pytest.raises(cvs_mod.CvsCommandException):
            self.reference_obj._cvs_update(filename,
                                           revision=revision,
                                           datetime_rev=revision_date)
        mock_cvs_command.assert_not_called()

    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_cvs_command')
    def test_cvs_update_negative(self, mock_cvs_command):
        mock_cvs_command.side_effect = cvs_mod.CvsCommandException('abc')

        with pytest.raises(cvs_mod.CvsCommandException):
            self.reference_obj._cvs_update()
        mock_cvs_command.assert_called_with('update -dP', exclusive=True)

    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_parse_cvs_update_output')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_cvs_checkout')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_cvs_update')
    def test_update_shared_repo(self,
                                mock_cvs_update,
                                mock_cvs_checkout,
                                mock_parse_cvs_output):
        mock_parse_cvs_output.return_value = []
        mock_cvs_update.return_value = ('std-out', 'std-err')

        self.reference_obj._update_shared_repo()
        mock_cvs_update.assert_called()

    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_clean_checkout_directory')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_parse_cvs_update_output')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_cvs_checkout')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_cvs_update')
    def test_update_shared_repo_with_cleanup(self,
                                             mock_cvs_update,
                                             mock_cvs_checkout,
                                             mock_parse_cvs_output,
                                             mock_clean_checkout_directory):
        mock_parse_cvs_output.side_effect = [
            ['clean/up/me/'],
            []
        ]
        mock_cvs_update.return_value = ('std-out', 'std-err')

        self.reference_obj._update_shared_repo()
        mock_cvs_update.assert_called()
        mock_clean_checkout_directory.assert_called()

    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_clean_checkout_directory')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_parse_cvs_update_output')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_cvs_checkout')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_cvs_update')
    def test_update_shared_repo_full_checkout(self,
                                              mock_cvs_update,
                                              mock_cvs_checkout,
                                              mock_parse_cvs_output,
                                              mock_clean_checkout_directory):
        mock_parse_cvs_output.side_effect = [
            ['clean/up/me/'],
            ['and/me/as/well']
        ]
        mock_cvs_update.return_value = ('std-out', 'std-err')

        self.reference_obj._update_shared_repo()
        mock_cvs_update.assert_called()
        mock_clean_checkout_directory.assert_called()
        mock_cvs_checkout.assert_called()

    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_clean_checkout_directory')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_parse_cvs_update_output')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_cvs_checkout')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_cvs_update')
    def test_update_shared_repo_cvs_failure(self,
                                            mock_cvs_update,
                                            mock_cvs_checkout,
                                            mock_parse_cvs_output,
                                            mock_clean_checkout_directory):
        mock_cvs_update.side_effect = cvs_mod.CvsCommandException('abc')

        self.reference_obj._update_shared_repo()

        mock_cvs_checkout.assert_called()

    def test_parse_cvs_update_output(self):
        cvsoutput = """
? rhel-unknown/vacations.tji
cvs update: warning: rhel-updated-wo-flag/Makefile was lost
U rhel-updated/Makefile
P rhel-patched/rhel-1-0-0.tjp
A rhel-added/rhel-1-0-0.tjp
A rhel-added/rhel-2-0-0.tjp
A rhel-added2/rhel-22-0-0.tjp
C rhel-conflict/rhel-2-0-0.tjp
M rhel-modified/rhel-3-0-0.tjp
x rhel-unknown-flag/rhel-3-0-0.tjp
"""
        to_cleanup = self.reference_obj._parse_cvs_update_output(cvsoutput)
        to_cleanup_ref = ['rhel-unknown', 'rhel-conflict', 'rhel-modified',
                          'rhel-unknown-flag', 'rhel-added', 'rhel-added2']
        assert to_cleanup == set(to_cleanup_ref)

    @mock.patch('schedules_tools.storage_handlers.cvs.os.path.exists')
    @mock.patch('schedules_tools.storage_handlers.cvs.remove_tree')
    def test_clean_checkout_directory(self, mock_remove_tree, mock_path_exists):
        directory = 'program/fedora/f-25'
        mock_path_exists.return_value = True
        self.reference_obj._clean_checkout_directory(directory)
        mock_remove_tree.assert_called_with('/tmp/mycheckoutdir/program/fedora/f-25')

    @mock.patch('schedules_tools.storage_handlers.cvs.os.path.exists')
    @mock.patch('schedules_tools.storage_handlers.cvs.remove_tree')
    def test_clean_not_existing_checkout_directory(self, mock_remove_tree, mock_path_exists):
        directory = 'program/fedora/f-25'
        mock_path_exists.return_value = False
        self.reference_obj._clean_checkout_directory(directory)
        mock_remove_tree.assert_not_called()

    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_cvs_command')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, 'refresh_local')
    def test_get_handle_changelog_cvs_fail(self,
                                           mock_refresh_local,
                                           mock_cvs_command):
        mock_cvs_command.side_effect = cvs_mod.CvsCommandException('abc')

        with pytest.raises(cvs_mod.CvsCommandException):
            self.reference_obj.get_handle_changelog()
        mock_refresh_local.assert_called()

    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_cvs_command')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, 'refresh_local')
    def test_get_handle_changelog(self,
                                  mock_refresh_local,
                                  mock_cvs_command):
        with open(os.path.join(BASE_DIR, 'fixtures', 'cvs_log.stdout')) as fd:
            cvs_stdout = fd.read()

        with open(os.path.join(BASE_DIR, 'fixtures', 'cvs_log.json')) as fd:
            changelog_ref = jsondate.load(fd)

        mock_cvs_command.return_value = (cvs_stdout, 'stderr')

        changelog = self.reference_obj.get_handle_changelog()
        assert changelog_ref == changelog
        mock_refresh_local.assert_called()

    @mock.patch('redis.StrictRedis')
    def test_cvs_command_exclusive_access_default_redis_config(
            self,
            mock_redis):
        options = {
            'cvs_repo_name': 'program',
            'cvs_root': ':gserver:someuser@cvs.myserver.com:/cvs/reporoot',
            'cvs_checkout_path': '/tmp/a/b/cc',
            'exclusive_access': True,
        }
        handle = 'program/product/release/release.tjp'
        reference = cvs_mod.StorageHandler_cvs(
            handle,
            options=options)
        mock_redis.assert_called_with()

    @mock.patch('redis.StrictRedis.from_url')
    def test_cvs_command_exclusive_access_custom_redis_config(
            self,
            mock_redis_from_url):
        options = {
            'cvs_repo_name': 'repo',
            'cvs_root': ':gserver:someuser@cvs.myserver.com:/cvs/reporoot',
            'cvs_checkout_path': '/tmp/a/b/cc',
            'exclusive_access': True,
            'redis_url': 'redis://redishost:1234/5'
        }
        handle = 'program/product/release/release.tjp'
        reference = cvs_mod.StorageHandler_cvs(
            handle,
            options=options)
        mock_redis_from_url.assert_called_with('redis://redishost:1234/5')

    @mock.patch('redis.StrictRedis')
    def test_cvs_command_exclusive_access_incorrect_custom_redis_config(
            self,
            mock_redis):
        options = {
            'cvs_repo_name': 'repo',
            'cvs_root': ':gserver:someuser@cvs.myserver.com:/cvs/reporoot',
            'cvs_checkout_path': '/tmp/a/b/cc',
            'exclusive_access': True,
            'cvs_lock_redis_uri': 'http://redishost:1234/5'
        }
        handle = 'program/product/release/release.tjp'
        reference = cvs_mod.StorageHandler_cvs(
            handle,
            options=options)
        mock_redis.assert_called_with()

    @mock.patch('redis.StrictRedis')
    @mock.patch('schedules_tools.storage_handlers.cvs.subprocess.Popen')
    def test_cvs_command_exclusive_access(self,
                                          mock_popen,
                                          mock_redis):
        options = {
            'cvs_repo_name': 'root',
            'cvs_root': ':gserver:someuser@cvs.myserver.com:/cvs/reporoot',
            'cvs_checkout_path': '/tmp/a/b/cc',
            'exclusive_access': True,
        }
        handle = 'program/product/release/release.tjp'

        cmd = 'update program/rhel'
        mock_communicate = mock.Mock()
        mock_communicate.communicate.return_value = ('stdo', 'stde')
        mock_popen.return_value = mock_communicate
        type(mock_communicate).returncode = mock.PropertyMock(return_value=0)

        mock_redis_inst = mock.Mock()
        mock_redis.return_value = mock_redis_inst

        mock_lock = mock.Mock()
        mock_redis_inst.lock.return_value = mock_lock

        reference = cvs_mod.StorageHandler_cvs(
            handle,
            options=options)
        assert reference._cvs_command(cmd, exclusive=True) == ('stdo', 'stde')
        mock_popen.assert_called()
        mock_lock.acquire.assert_called()
        mock_lock.release.assert_called()

    @mock.patch('redis.StrictRedis')
    def test_cvs_command_exclusive_access_unable_acquire(self, mock_redis):
        options = {
            'cvs_repo_name': 'repo',
            'cvs_root': ':gserver:someuser@cvs.myserver.com:/cvs/reporoot',
            'cvs_checkout_path': '/tmp/a/b/cc',
            'exclusive_access': True,
        }
        handle = 'program/product/release/release.tjp'

        cmd = 'update program/rhel'

        mock_redis_inst = mock.Mock()
        mock_redis.return_value = mock_redis_inst

        mock_acquire = mock.Mock(return_value=False)
        mock_lock = mock.Mock()
        mock_lock.acquire = mock_acquire
        mock_redis_inst.lock.return_value = mock_lock

        reference = cvs_mod.StorageHandler_cvs(
            handle,
            options=options)

        with pytest.raises(storage_handlers.AcquireLockException):
            reference._cvs_command(cmd, exclusive=True)

    @mock.patch('shutil.move')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_wipe_out_dir')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_is_valid_cvs_dir')
    def test_checkout_dir_by_opt_arg(self,
                                     mock_is_valid_cvs_dir,
                                     mock_wipe_out_dir,
                                     mock_shutil_move):
        options = {
            'cvs_repo_name': 'repo',
            'cvs_root': ':gserver:someuser@cvs.myserver.com:/cvs/reporoot',
            'cvs_checkout_path': '/tmp/a/b/cc'
        }
        handle = 'program/product/release/release.tjp'
        reference = cvs_mod.StorageHandler_cvs(
            handle,
            options=options)

        assert reference._cvs_checkout() == '/tmp/a/b/cc'

    @mock.patch('shutil.move')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_wipe_out_dir')
    @mock.patch('tempfile.mkdtemp')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_cvs_command')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_is_valid_cvs_dir')
    def test_checkout_dir_by_opt_arg_force(self,
                                           mock_is_valid_cvs_dir,
                                           mock_cvs_command,
                                           mock_mkdtemp,
                                           mock_wipe_out_dir,
                                           mock_shutil_move):
        options = {
            'cvs_repo_name': 'repo',
            'cvs_root': ':gserver:someuser@cvs.myserver.com:/cvs/reporoot',
            'cvs_checkout_path': '/tmp/somepath',
        }
        handle = 'program/product/release/release.tjp'
        reference = cvs_mod.StorageHandler_cvs(
            handle,
            options=options)

        mock_is_valid_cvs_dir.return_value = True
        mock_cvs_command.return_value = ('stdout', 'stderr')

        assert reference._cvs_checkout(force=True) == '/tmp/somepath'
        mock_mkdtemp.assert_called()
        mock_wipe_out_dir.assert_called()
        mock_shutil_move.assert_called()
        mock_cvs_command.assert_called()

    @mock.patch('shutil.move')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_wipe_out_dir')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_cvs_command')
    @mock.patch.object(cvs_mod.StorageHandler_cvs, '_is_valid_cvs_dir')
    @mock.patch('tempfile.mkdtemp')
    def test_checkout_dir_mkdtemp(self,
                                  mock_mkdtemp,
                                  mock_is_valid_cvs_dir,
                                  mock_cvs_command,
                                  mock_wipe_out_dir,
                                  mock_shutil_move):
        options = {
            'cvs_repo_name': 'repo',
            'cvs_root': ':gserver:someuser@cvs.myserver.com:/cvs/reporoot',
            'cvs_checkout_path': None,
        }
        handle = 'program/product/release/release.tjp'
        reference = cvs_mod.StorageHandler_cvs(handle, options=options)

        mock_mkdtemp.side_effect = ['/temp/repo', '/temp/temp_cloed_repo']

        mock_is_valid_cvs_dir.return_value = True

        mock_cvs_command.return_value = ('stdout', 'stderr')

        assert reference._cvs_checkout() == '/temp/repo'
        mock_mkdtemp.assert_called()
        mock_wipe_out_dir.assert_called()
        mock_shutil_move.assert_called()



from schedules_tools.schedule_handlers import ScheduleHandlerBase
from schedules_tools import models
import sys
import datetime
import os
import logging

log = logging.getLogger(__name__)

try:
    from pyral import Rally, rallySettings
    additional_deps_satistifed = True
except ImportError:
    additional_deps_satistifed = False


class ScheduleHandler_rally(ScheduleHandlerBase):
    provide_export = False

    handle_deps_satisfied = additional_deps_satistifed

    @classmethod
    def is_valid_source(cls, handle=None):
        if not handle:
            handle = cls.handle
        if os.stat(handle).st_size < 1024:
            file_cont = open(handle).read()
            if 'WORKSPACE' in file_cont and 'PROJECT' in file_cont:
                return True
        return False

    def import_schedule(self):
        self.schedule = models.Schedule()
        start_time = None

        options = ['--config=%s' % self.handle]
        server, user, password, workspace, project = rallySettings(options)

        rally = Rally(server, user, password, workspace=workspace, project=project)

        rally_iter = self.options['rally_iter']
        self.schedule.name = rally_iter.strip()
        query_criteria = 'Iteration.Name = "%s"' % rally_iter

        response = rally.get('Iteration', fetch=True,
                             query='Name = "%s"' % rally_iter)
        if response.errors:
            sys.stdout.write("\n".join(response.errors))
            sys.exit(1)

        for iteration in response:
            print 'Iteration: %s (starts %s)' % (iteration.Name, iteration.StartDate)
            start_time = datetime.datetime.combine(
                datetime.datetime.strptime(iteration.StartDate[:10], '%Y-%m-%d'),
                datetime.time(8))
            break

        response = rally.get('UserStory', fetch=True, query=query_criteria, order="Rank")

        if response.errors:
            sys.stdout.write("\n".join(response.errors))
            sys.exit(1)

        index = 1
        if not start_time:
            start_time = datetime.datetime.combine(datetime.date.today(), datetime.time(8))
        max_end_time = start_time
        self.schedule.dStart = start_time

        for story in response:
            print story.Name
            t = models.Task(self.schedule, level=1)
            t.index = index
            index += 1
            t.name = story.Name.strip()
            t.dStart = start_time

            max_st_end_time = start_time
            story.Tasks.sort(key=lambda x: x.TaskIndex)
            for task in story.Tasks:
                print '-- %s  |  %sh  |  %s' % (task.Name, task.Estimate, task.Owner.Name)
                t_in = models.Task(self.schedule, level=2)
                t_in.index = index
                index += 1
                t_in.name = task.Name.strip()
                t_in.dStart = start_time
                t_in.dFinish = start_time + datetime.timedelta(hours=float(task.Estimate))
                max_st_end_time = max(max_end_time, t_in.dFinish)

                # look for resource
                resource_id = None
                for r_id, resource in self.schedule.resources.items():
                    if resource == task.Owner.Name:
                        resource_id = r_id
                        break
                if not resource_id:
                    resource_id = len(self.schedule.resources) + 1
                    self.schedule.resources[resource_id] = str(task.Owner.Name)

                t_in.resource = resource_id

                t.tasks.append(t_in)
            print ''
            t.dFinish = max_st_end_time
            max_end_time = max(max_end_time, t.dFinish)
            self.schedule.tasks.append(t)

        self.schedule.dFinish = max_end_time
        return self.schedule




import os
import tempfile
import logging

from datetime import datetime
from datetime import timedelta
from lxml import etree

from schedules_tools import models, SchedulesToolsException
from schedules_tools.schedule_handlers import ScheduleHandlerBase


MSP_FLAGS_ATTRS = ('Flags', )
datetime_format = '%Y-%m-%dT%H:%M:%S'

log = logging.getLogger(__name__)


class MSPImportException(SchedulesToolsException):
    pass


class ScheduleHandler_msp(ScheduleHandlerBase):
    provide_export = True

    handle_deps_satisfied = True
    
    default_export_ext = 'xml'

    # amount of working hours per day
    working_hours = 8

    @classmethod
    def is_valid_source(cls, handle=None):
        if not handle:
            handle = cls.handle
        try:
            tree = etree.parse(handle)
        except (etree.XMLSyntaxError, IOError):
            return False

        if 'http://schemas.microsoft.com/project' in tree.getroot().tag:
            return True
        return False

    # Schedule
    def import_schedule(self):
        self.schedule = models.Schedule()

        try:
            # remove project's xmlns
            tmp_file = tempfile.mkstemp()[1]
            with open(tmp_file, 'wt') as hTmp_file:
                for line in open(self.handle):
                    hTmp_file.write(line.replace(' xmlns="http://schemas.microsoft.com/project"', ''))
    
            start_level = 1
            tree = etree.parse(tmp_file)
    
            eTask_list = tree.xpath('Tasks/Task[OutlineLevel >= %s]' % start_level)
            self.schedule.name = tree.xpath('Name|Title')[0].text.strip()
            self.schedule.slug = self.schedule.unique_id_re.sub('_', self.schedule.name)
    
            # extended attributes
            for eExtAttr in tree.xpath('ExtendedAttributes/ExtendedAttribute'):
                fieldID = int(eExtAttr.xpath('FieldID')[0].text)
                fieldName = eExtAttr.xpath('FieldName')[0].text
                self.schedule.ext_attr[fieldName] = fieldID
    
            if self.schedule.ext_attr:  # choose flags field
                for ff_name in MSP_FLAGS_ATTRS:
                    if ff_name in self.schedule.ext_attr:
                        self.schedule.flags_attr_id = self.schedule.ext_attr[ff_name]
                        break
    
            self.schedule.tasks = self._load_tasks_level(start_level, eTask_list)
    
            os.unlink(tmp_file)
            self.schedule.check_top_task()
            self.schedule.generate_slugs()
            return self.schedule
        except etree.XMLSyntaxError as e:
            raise MSPImportException(e, source=self.handle)

    # Schedule
    def export_schedule(self, out_file=None):
        MSP_NAMESPACE = "http://schemas.microsoft.com/project"
        MSP = "{%s}" % MSP_NAMESPACE

        NSMAP = {None: MSP_NAMESPACE}  # the default namespace (no prefix)

        eProject = etree.Element(MSP + 'Project', nsmap=NSMAP)
        eName = etree.SubElement(eProject, 'Name')
        eName.text = self.schedule.name
        eTitle = etree.SubElement(eProject, 'Title')
        eTitle.text = self.schedule.name
        eMinutesPerDay = etree.SubElement(eProject, 'MinutesPerDay')
        working_minutes_per_day = self.working_hours * 60
        eMinutesPerDay.text = str(working_minutes_per_day)
        eMinutesPerWeek = etree.SubElement(eProject, 'MinutesPerWeek')
        # 7 working days
        eMinutesPerWeek.text = str(working_minutes_per_day * 7)

        calendar_UID = '1'
        # add ext condition
        eCalendarUID = etree.SubElement(eProject, 'CalendarUID')
        eCalendarUID.text = calendar_UID
        eCalendars = etree.SubElement(eProject, 'Calendars')
        eCalendar = etree.SubElement(eCalendars, 'Calendar')
        eUID = etree.SubElement(eCalendar, 'UID')
        eUID.text = calendar_UID
        eName = etree.SubElement(eCalendar, 'Name')
        eName.text = 'Standard'
        eIsBaseCalendar = etree.SubElement(eCalendar, 'IsBaseCalendar')
        eIsBaseCalendar.text = '1'
        eWeekDays = etree.SubElement(eCalendar, 'WeekDays')
        for day_type in range(1, 8):
            eWeekDay = etree.SubElement(eWeekDays, 'WeekDay')
            eDayType = etree.SubElement(eWeekDay, 'DayType')
            eDayType.text = str(day_type)
            eDayWorking = etree.SubElement(eWeekDay, 'DayWorking')
            eDayWorking.text = '1'
            eWorkingTimes = etree.SubElement(eWeekDay, 'WorkingTimes')

            # Sunday and Saturday don't have specified working hours
            if day_type in [1, 7]:
                continue

            for from_time in [(9, 3), (13, 5)]:
                eWorkingTime = etree.SubElement(eWorkingTimes, 'WorkingTime')
                eFromTime = etree.SubElement(eWorkingTime, 'FromTime')
                eFromTime.text = '%02d:00:00' % int(from_time[0])
                eToTime = etree.SubElement(eWorkingTime, 'ToTime')
                eToTime.text = '%02d:00:00' % (from_time[0] + from_time[1],)

        for r_id, resource in self.schedule.resources.items():
            eCalendar = etree.SubElement(eCalendars, 'Calendar')
            eUID = etree.SubElement(eCalendar, 'UID')
            eUID.text = str(r_id + 1)
            eName = etree.SubElement(eCalendar, 'Name')
            eName.text = resource
            eIsBaseCalendar = etree.SubElement(eCalendar, 'IsBaseCalendar')
            eIsBaseCalendar.text = '0'
            eBaseCalendarUID = etree.SubElement(eCalendar, 'BaseCalendarUID')
            eBaseCalendarUID.text = calendar_UID

        # SmartSheet workaround:
        # It's necessary to define at least one non-working days, due bug
        # in SmartSheet import, otherwise whole calendar is ignored..
        eExceptions = etree.SubElement(eCalendar, 'Exceptions')
        eException = etree.SubElement(eExceptions, 'Exception')
        eTimePeriod = etree.SubElement(eException, 'TimePeriod')
        eFromDate = etree.SubElement(eTimePeriod, 'FromDate')
        eToDate = etree.SubElement(eTimePeriod, 'ToDate')
        # use date 5y before whole schedule beginning
        exception_from_date = self.schedule.dStart.replace(
            year=self.schedule.dStart.year - 5)
        exception_to_date = exception_from_date + timedelta(days=1)
        eFromDate.text = exception_from_date.strftime(datetime_format)
        eToDate.text = exception_to_date.strftime(datetime_format)
        eType = etree.SubElement(eException, 'Type')
        eType.text = '1'

        eTasks = etree.SubElement(eProject, 'Tasks')
        self.export_msp_tasks(self.schedule.tasks, eTasks, '')

        eResources = etree.SubElement(eProject, 'Resources')
        for r_id, resource in self.schedule.resources.items():
            eResource = etree.SubElement(eResources, 'Resource')

            eUID = etree.SubElement(eResource, 'UID')
            eUID.text = str(r_id)

            eID = etree.SubElement(eResource, 'ID')
            eID.text = str(r_id)

            eName = etree.SubElement(eResource, 'Name')
            eName.text = resource

            eCalendarUID = etree.SubElement(eResource, 'CalendarUID')
            eCalendarUID.text = str(r_id + 1)

        eAssignments = etree.SubElement(eProject, 'Assignments')
        for assignment in self.schedule.assignments:
            eAssignment = etree.SubElement(eAssignments, 'Assignment')

            eTaskUID = etree.SubElement(eAssignment, 'TaskUID')
            eTaskUID.text = str(assignment['t_id'])

            eResourceUID = etree.SubElement(eAssignment, 'ResourceUID')
            eResourceUID.text = str(assignment['r_id'])

            eUnits = etree.SubElement(eAssignment, 'Units')
            eUnits.text = '1'

        et = etree.ElementTree(eProject)
        
        if out_file:
            et.write(out_file, pretty_print=True, encoding="utf-8",
                     xml_declaration=True)
        
        return str(et)        

    # Schedule
    def export_msp_tasks(self, tasks, eParent, outline_prefix):
        for n, task in enumerate(tasks, start=1):
            eTask = self.task_export_msp_node(task)
            eUID = etree.SubElement(eTask, 'UID')
            eUID.text = str(self.schedule._task_index)

            eID = etree.SubElement(eTask, 'ID')
            eID.text = str(self.schedule._task_index)

            if task.resource:
                self.schedule.assignments.append({
                    't_id': self.schedule._task_index,
                    'r_id': task.resource}
                )

            eOutlineNumber = etree.SubElement(eTask, 'OutlineNumber')
            eOutlineNumber.text = '%s%s' % (outline_prefix, n)

            eOutlineLevel = etree.SubElement(eTask, 'OutlineLevel')
            eOutlineLevel.text = str(len(outline_prefix.split('.')))

            eParent.append(eTask)

            self.schedule._task_index += 1
            self.export_msp_tasks(task.tasks, eParent,
                                  '%s.' % eOutlineNumber.text)

    # Schedule
    def _load_tasks_level(self, level, eTask_list):
        return_tasks = []

        while eTask_list:
            eTask = eTask_list[0]
            task_level = int(eTask.xpath('OutlineLevel')[0].text)

            if task_level > level:
                # return tasks may be empty since there could be no importable tasks yet
                if len(return_tasks):
                    return_tasks[-1].tasks = self._load_tasks_level(
                        task_level, eTask_list)
                else:
                    # remove task from list
                    eTask_list.pop(0)
                continue
            elif task_level < level:
                return return_tasks

            # process task
            task = models.Task(self.schedule, level=level)
            if self.task_load_msp_node(task, eTask):                
                # update schedule start/end
                if self.schedule.dStart:
                    self.schedule.dStart = min(self.schedule.dStart, task.dStart)
                else:
                    self.schedule.dStart = task.dStart

                if self.schedule.dFinish:
                    self.schedule.dFinish = max(self.schedule.dFinish, task.dAcFinish)
                else:
                    self.schedule.dFinish = task.dAcFinish
                    
                return_tasks.append(task)
            # remove task from list
            eTask_list.pop(0)
        return return_tasks

    # Task
    def task_load_msp_node(self, task, eTask):
        task.index = int(eTask.xpath('ID')[0].text)

        task.name = task._workaround_it_phase_names(eTask)
        if not task.name:
            return False

        task.priority = int(eTask.xpath('Priority')[0].text)

        nlStart = eTask.xpath('Start')

        if nlStart:
            task.dStart = task.dAcStart = datetime.strptime(nlStart[0].text,
                                                   task._date_format)
            task.dFinish = task.dAcFinish = task.dStart
        else:
            return False

        nlFinish = eTask.xpath('Finish')
        if nlFinish:
            task.dFinish = task.dAcFinish = datetime.strptime(
                nlFinish[0].text, task._date_format)

        nlAcStart = eTask.xpath('ActualStart')
        if nlAcStart:
            task.dAcStart = datetime.strptime(nlAcStart[0].text,
                                              task._date_format)

        nlAcFinish = eTask.xpath('ActualFinish')
        if nlAcFinish:
            task.dAcFinish = datetime.strptime(nlAcFinish[0].text,
                                               task._date_format)

        # sanity check - if only ac start defined and beyond plan finish
        task.dAcFinish = max(task.dAcFinish, task.dAcStart)

        task.milestone = eTask.xpath('Milestone')[0].text == '1'

        ePercentComplete_list = eTask.xpath('PercentComplete')
        if ePercentComplete_list:
            task.p_complete = float(eTask.xpath('PercentComplete')[0].text)

        notes = eTask.xpath('Notes')
        if notes:
            task.note = notes[0].text.strip()

        # load flags from ext attributes
        flag_ext_attr = eTask.xpath(
            'ExtendedAttribute[FieldID = {}]'.format(
                task._schedule.flags_attr_id)
        )
        if flag_ext_attr:
            flags_value = flag_ext_attr[0].xpath('Value')[0].text
            if flags_value:
                task.flags = [f for f in flags_value.strip(' ,\n').split(',') if ' ' not in f]
                task._schedule.used_flags |= set(task.flags)

        # workaround for SmartSheet exports - load flags, links
        # @param value: XPath element instance (/Project/Tasks/Task/ExtendedAttribute/Value)
        ext_attr_elements = eTask.xpath('ExtendedAttribute/Value')
        for ext_attr in ext_attr_elements:
            task.parse_extended_attr(ext_attr.text)

        task.check_for_phase()
        return True

    # Task
    def task_export_msp_node(self, task):
        eTask = etree.Element('Task')

        eName = etree.SubElement(eTask, 'Name')
        eName.text = task.name

        eStart = etree.SubElement(eTask, 'Start')
        eStart.text = task.dStart.strftime(datetime_format)
        eFinish = etree.SubElement(eTask, 'Finish')
        eFinish.text = task.dFinish.strftime(datetime_format)

        ePriority = etree.SubElement(eTask, 'Priority')
        ePriority.text = str(task.priority)

        # valid content is 0 or 1, not 'True', 'False'
        eMilestone = etree.SubElement(eTask, 'Milestone')
        eMilestone.text = str(int(task.milestone))

        duration = task.dFinish - task.dStart
        # Workaround for SmartSheet import:
        # They consider duration always +1 greater, than is difference between
        # finish and start date in whole days, Finish element is ignored by them
        duration += timedelta(days=1)
        eDuration = etree.SubElement(eTask, 'Duration')
        h, rem = divmod(duration.seconds, 3600)
        h += duration.days * self.working_hours
        m, s = divmod(rem, 60)
        eDuration.text = 'PT%sH%sM%sS' % (h, m, s)
        eDurationFormat = etree.SubElement(eTask, 'DurationFormat')
        eDurationFormat.text = '39'

        if task.note:
            eNotes = etree.SubElement(eTask, 'Notes')
            eNotes.text = task.note

        flags_str = ','.join(task.flags)
        if flags_str:
            ext_attr_element = etree.SubElement(eTask, 'ExtendedAttribute')
            value_element = etree.SubElement(ext_attr_element, 'Value')
            value_element.text = '{}: {}'.format(models.ATTR_PREFIX_FLAG, flags_str)

            # this value is not used, but required by SmartSheets import
            fieldid_element = etree.SubElement(ext_attr_element, 'FieldID')
            fieldid_element.text = '188743734'
        if task.link:
            ext_attr_element = etree.SubElement(eTask, 'ExtendedAttribute')
            value_element = etree.SubElement(ext_attr_element, 'Value')
            value_element.text = '{}: {}'.format(models.ATTR_PREFIX_LINK, task.link)

            # this value is not used, but required by SmartSheets import
            fieldid_element = etree.SubElement(ext_attr_element, 'FieldID')
            fieldid_element.text = '188743737'

        return eTask



from schedules_tools.schedule_handlers import ScheduleHandlerBase
import logging
from lxml.html import etree

log = logging.getLogger(__name__)


css = """
a[href=""] {display:none}

table.schedule {
    border-collapse: collapse;
}

table.schedule th, table.schedule td {
    border: 2px solid black;
    padding: 3px 5px;
}

table.schedule th {
    background-color: #a5c2ff;
}
table.schedule td {
    background-color: #f3ebae;
}
table.schedule td.date {
    font-size: 90%;
    white-space: nowrap;
    text-align: right;
}
table.schedule td.duration {
    text-align: right;
}

table.schedule td div.note {
    font-size: 80%;
}
"""


class ScheduleHandler_html(ScheduleHandlerBase):
    provide_export = True

    handle_deps_satisfied = True

    default_export_ext = 'html'

    indent_level_px = 14

    def __init__(self, *args, **kwargs):
        super(ScheduleHandler_html, self).__init__(*args, **kwargs)
        if not self.options.get('date_format', False):
            self.options['date_format'] = '%a %Y-%m-%d'

    @classmethod
    def is_valid_source(cls, handle=None):
        return False

    def _export_task(self, e_table, task, hiearchy_parent='',
                     hiearchy_index=''):
        e_tr = etree.SubElement(e_table, 'tr')
        e_td = etree.SubElement(e_tr, 'td')
        curr_hiearchy_index = str(hiearchy_parent)
        if hiearchy_index:
            curr_hiearchy_index += '.' + str(hiearchy_index)
        e_td.text = curr_hiearchy_index

        padding = (task.level - 1) * self.indent_level_px
        e_td = etree.SubElement(e_tr, 'td',
                                style='padding-left: {}px'.format(padding))
        e_td.text = task.name

        if task.note:
            e_note = etree.SubElement(e_td, 'div')
            e_note.attrib['class'] = 'note'
            e_note.text = task.note

        if task.link:
            e_div = etree.SubElement(e_td, 'div')
            e_link = etree.SubElement(e_div, 'a')
            e_link.attrib['href'] = task.link
            e_link.text = task.link

        e_td = etree.SubElement(e_tr, 'td')
        e_td.attrib['class'] = 'date'
        e_td.text = str(task.dAcStart.strftime(self.options['date_format']))

        e_td = etree.SubElement(e_tr, 'td')
        e_td.attrib['class'] = 'date'
        e_td.text = str(task.dAcFinish.strftime(self.options['date_format']))

        duration = task.dAcFinish - task.dAcStart
        e_td = etree.SubElement(e_tr, 'td')
        e_td.attrib['class'] = 'duration'
        e_td.text = str(duration.days)

        for index, task in enumerate(task.tasks):
            self._export_task(e_table, task, curr_hiearchy_index, index + 1)

    # Schedule
    def export_schedule(self, out_file=None):
        e_html = etree.Element('html')
        e_head = etree.SubElement(e_html, 'head')

        e_encoding = etree.SubElement(e_head, 'meta', charset="utf-8")

        if self.options.get('html_title', False):
            title = self.options['html_title']
        else:
            title = self.schedule.name
                    
        e_title = etree.SubElement(e_head, 'title')
        e_title.text = title

        e_style = etree.SubElement(e_head, 'style', type='text/css')
        e_style.text = css

        e_body = etree.SubElement(e_html, 'body')

        e_h1 = etree.SubElement(e_body, 'h1')
        e_h1.text = title
        
        if self.options.get('html_table_header', False):
            e_body.append(etree.fromstring(self.options['html_table_header']))

        e_table = etree.SubElement(e_body, 'table', attrib={'align': 'center',
                                                            'class': 'schedule'})
        e_tr_head = etree.SubElement(e_table, 'tr')
        head_columns = ['HierarchIndex', 'Name', 'Start', 'End', 'Duration']
        for column in head_columns:
            e_th_head = etree.SubElement(e_tr_head, 'th')
            e_th_head.text = column

        for index, task in enumerate(self.schedule.tasks):
            self._export_task(e_table, task, index + 1)

        etree_return = etree.ElementTree(e_html)
        if out_file:
            etree_return.write(out_file, pretty_print=True, encoding="utf-8",
                               xml_declaration=False)
        
        return str(etree_return)



import json
import re
import logging

from schedules_tools import jsondate
from schedules_tools.models import Task
from deepdiff import DeepDiff
from deepdiff.helper import NotPresent

log = logging.getLogger(__name__)


def _custom_json_encoder(obj):
    if isinstance(obj, set):
        return list(obj)
    return jsondate._datetime_encoder(obj)


REPORT_NO_CHANGE = ''
REPORT_ADDED = '_added_'
REPORT_REMOVED = '_removed_'
REPORT_CHANGED = '_changed_'
REPORT_NEW_VALUE = 'new_value'
REPORT_OLD_VALUE = 'old_value'

REPORT_KEYS = {REPORT_ADDED, REPORT_REMOVED, REPORT_CHANGED, }

REPORT_PREFIX_MAP = {
    REPORT_ADDED: '[+]',
    REPORT_REMOVED: '[-]',
    REPORT_CHANGED: '[M]',
    REPORT_NO_CHANGE: 3 * ' ',
}


class ScheduleDiff(object):

    def __init__(self, schedule_a, schedule_b):
        self.schedule_a = schedule_a
        self.schedule_b = schedule_b

        self.schedule_a_dict = schedule_a.dump_as_dict()
        self.schedule_b_dict = schedule_b.dump_as_dict()

        self.diff = DeepDiff(self.schedule_a_dict, self.schedule_b_dict,
                             verbose_level=0,  # ignore type changes
                             view='tree'
                             )

    def filter_attrs(self, in_dict, attrs):
        return {k: v for k, v in in_dict.iteritems() if k in attrs}

    def walk_dict(self, in_dict, keys_list):
        """ Returns the value of the given keys_list path in the dictionary. """

        if len(keys_list) > 0:
            key = keys_list.pop(0)

            if isinstance(in_dict, list):
                item = in_dict[int(key)]
            else:
                item = in_dict[key]

            in_dict = self.walk_dict(item, keys_list)
        return in_dict

    def path_to_keys(self, path):
        """
        Transforms a path str into list of keys,
        collecting only the values wrapped with brackets.

        Example: "root['attr1'][1]" becomes ['attr1', '1']
        """
        return re.findall(r"\[[']*([\w.]+)[']*\]", path)

    def _create_change_report(self, change_type, new_value, old_value):
        """
        Returns a dictionary with the change report.

        @param change_type: "added", "removed" or "changed"
        """
        change_value = {}
        change_key = '_%s_' % change_type

        if change_type == 'changed':
            if not isinstance(new_value, NotPresent):
                change_value['new_value'] = new_value
            if not isinstance(old_value, NotPresent):
                change_value['old_value'] = old_value

        else:  # added or removed
            change_value = new_value if change_type == 'added' else old_value

        return {change_key: change_value}

    def _apply_change_to_tasks(self, tasks, change_type):
        res = []

        for task in tasks:
            report = self._create_change_report(change_type,
                                                {} if change_type == 'removed' else task,
                                                {} if change_type == 'added' else task)

            report_value = report.itervalues().next()

            if task['tasks']:
                report_value['tasks'] = self._apply_change_to_tasks(task['tasks'], change_type)

            res.append(report)
        return res

    def _add_change_report(self, to_dict, change):
        """
        Place a change report in the given dictionary, using the path of the change object.

        NOTE:

            Changes with paths not found in the given dictionary will be skipped.

        @param to_dict: dictionary to be updated with the change reports
        @param change: DiffLevel object that represents a change.
        """
        change_type = change.report_type.split('_')[-1]  # added, removed or changed
        path = self.path_to_keys(change.path())
        changed_key = path.pop()

        try:
            parent = self.walk_dict(to_dict, path)
        except (ValueError, KeyError):
            log.warning("Change skipped."
                           "Could not find any item with the path: %s" % change.path())
            return

        change_report = self._create_change_report(change_type,
                                                   new_value=change.t2,
                                                   old_value=change.t1)

        if isinstance(parent, list):
            index = int(changed_key)
            report_key, report_value = change_report.iteritems().next()

            if report_key == REPORT_ADDED:
                parent.insert(index, change_report)

            else:
                if 'tasks' in report_value:
                    report_value['tasks'] = self._apply_change_to_tasks(parent[index]['tasks'], change_type)

                parent[index] = change_report
        else:
            parent[changed_key] = change_report

    def dump_dict(self, schedule_attrs=[]):
        """
        Returns a dictionary that represents the result of the diff.

        @param  schedule_attrs: Set of schedule attrs to be returned.
                                If not specified, will return all attrs.
        """
        filter_attrs = schedule_attrs or self.schedule_a_dict.keys()
        diff_dict = self.filter_attrs(self.schedule_a_dict, filter_attrs)

        def add_to_dict(change):
            self._add_change_report(diff_dict, change)

        self._for_each_change(add_to_dict, schedule_attrs)

        return diff_dict

    def dump_json(self, schedule_attrs=[], **kwargs):
        """
        Serialize diff result to a JSON formatted str.

        @param  schedule_attrs: Set of schedule attrs to be returned.
                                If not specified, will return all attrs.
        @param  **kwargs: Can be used to pass in arguments to json.dumps().
        """
        kwargs['default'] = _custom_json_encoder
        res_dict = self.dump_dict(schedule_attrs)
        return json.dumps(res_dict, **kwargs)

    def _for_each_change(self, perform, filter_attrs=[]):
        """
        Iterates through the list of changes and performs the action passed in as argument.

        @param perform: Function to execute on each iteration.
        """
        for key, changes in self.diff.iteritems():
            for change in changes:
                path = change.path()
                sched_attr = self.path_to_keys(path)[0]

                if not filter_attrs or sched_attr in filter_attrs:
                    # execute the given function
                    perform(change)

    def _task_to_str(self, task, prefix='', level=0):
        task = Task.load_from_dict(task, schedule=None)
        return "{} {}{}".format(prefix, level * ' ', str(task))

    def _get_task_diff_str(self, task, change_type='', level=0):
        prefix = REPORT_PREFIX_MAP[REPORT_NO_CHANGE]

        if change_type in REPORT_KEYS:
            prefix = REPORT_PREFIX_MAP[change_type]

        return self._task_to_str(task, prefix=prefix, level=level)

    def _merge_task(self, task):
        merged_task = task
        change_type = ''

        for key, value in task.iteritems():
            if key in REPORT_KEYS:
                merged_task = value[REPORT_NEW_VALUE] if key == REPORT_CHANGED else value
                change_type = key
            elif self.contains_change_report(value):
                merged_task[key], change_type = self._merge_task(value)

        return merged_task, change_type

    def contains_change_report(self, item):
        return isinstance(item, dict) and any(key in item for key in REPORT_KEYS)

    def tasks(self, tasks=None, level=0):
        """ Textual representation of the tasks' diff. """
        res = ''

        if tasks is None:
            tasks_diff_dict = self.dump_dict(schedule_attrs=['tasks'])
            tasks = tasks_diff_dict['tasks']

        for task in tasks:
            merged_task, change_type = self._merge_task(task)
            res += '%s\n' % self._get_task_diff_str(merged_task, change_type, level)

            if merged_task['tasks']:
                res += self.tasks(merged_task['tasks'], level + 1)

        return res



from dateutil.tz import tzutc
import datetime
import json
import logging
import os
import pytest
import re
import shutil
from smartsheet import Smartsheet
from schedules_tools.schedule_handlers.smart_sheet import (
    SmartSheetExportException)


from schedules_tools.converter import ScheduleConverter
from schedules_tools.tests import jsondate
from schedules_tools.models import Schedule

# smartsheet log
logging.basicConfig(level=logging.INFO)

log = logging.getLogger(__name__)


def pytest_generate_tests(metafunc):
    source_function = metafunc.function.__name__

    if source_function == 'test_import':
        argnames = ['handler_name', 'import_schedule_file']
        argvalues = metafunc.cls.scenarios_import_combinations
        metafunc.parametrize(argnames, argvalues)
    elif source_function == 'test_export':
        argnames = ['handler_name', 'export_schedule_file', 'flat', 
                    'flag_show', 'flag_hide', 
                    'options']
        argvalues = metafunc.cls.scenarios_export_combinations
        metafunc.parametrize(argnames, argvalues)


class TestHandlers(object):
    intermediary_reference_file = 'intermediary-struct-reference.json'
    schedule_files_dir = 'schedule_files'
    basedir = os.path.dirname(os.path.realpath(__file__))
    test_import_start_timestamp = None

    scenarios_import_combinations = [
        ('msp', 'import-schedule-msp.xml'),
        ('smartsheet', ''),
        ('json', 'import-schedule-json.json'),
        ('tjx2', 'import-schedule-tjx2.tjx'),
    ]
    scenarios_export_combinations = [
        ('msp', 'export-schedule-msp.xml', False, [], [], {}),
        ('json', 'export-schedule-json.json', False, [], [], {}),
        ('json', 'export-schedule-json.json-flags', False, ['flag1'], ['flag2'], {}),
        ('json', 'export-schedule-json-flat.json', True, [], [], {}),
        ('json', 'export-schedule-json-flat-flags.json', True, ['flag1'], ['flag2'], {}),
        ('ics', 'export-schedule-ics.ics', False, [], [], {}),
        ('html', 'export-schedule-html.html', False, [], [], {}),
        ('html', 'export-schedule-html-options.html', False, [], [],
         dict(html_title='Test title', html_table_header='<p>Test header</p>')),
    ]

    smartsheet_columns_ids = ()
    smartsheet_sheet_id = None
    smartsheet_client = None
    replace_time_opts = dict(hour=0, minute=0, tzinfo=None)

    def _sanitize_export_test_ics(self, content):
        return re.sub('DTSTAMP:[0-9]+T[0-9]+Z', 'DTSTAMP:20170101T010101Z', content)

    def _clean_interm_struct(self, input_dict):
        """Removes keys that is not needed for comparison,
        unify time-part of dates"""
        keys_to_remove = ['unique_id_re', 'id_reg', 'ext_attr', 'flags_attr_id',
                          'resources', 'mtime']

        # remove schedule attrs
        for key in keys_to_remove:
            if key in input_dict:
                input_dict.pop(key)

        # Schedule attrs
        input_dict['dStart'] = input_dict['dStart'].replace(**self.replace_time_opts)
        input_dict['dFinish'] = input_dict['dFinish'].replace(**self.replace_time_opts)

        # phases cleanup
        for phase in input_dict['phases']:
            phase['dStart'] = phase['dStart'].replace(**self.replace_time_opts)
            phase['dFinish'] = phase['dFinish'].replace(**self.replace_time_opts)
            phase['dAcStart'] = phase['dAcStart'].replace(**self.replace_time_opts)
            phase['dAcFinish'] = phase['dAcFinish'].replace(**self.replace_time_opts)

        # Task(s) attrs
        for task in input_dict['tasks']:
            self._clear_task_time(task)

    def _clear_task_time(self, task):
        """For comparison purpose we ignore hours and minutes of task."""
        task['dStart'] = task['dStart'].replace(**self.replace_time_opts)
        task['dAcStart'] = task['dAcStart'].replace(**self.replace_time_opts)
        task['dFinish'] = task['dFinish'].replace(**self.replace_time_opts)
        task['dAcFinish'] = task['dAcFinish'].replace(**self.replace_time_opts)

        for inner_task in task['tasks']:
            self._clear_task_time(inner_task)

    @classmethod
    def _convert_struct_unicode_to_str(cls, data, ignore_dicts=False):
        # Taken from:
        # https://stackoverflow.com/questions/956867/how-to-get-string-objects-instead-of-unicode-from-json

        # if this is a unicode string, return its string representation
        if isinstance(data, unicode):
            return data.encode('utf-8')

        # if this is a list of values, return list of byteified values
        if isinstance(data, list):
            return [
                cls._convert_struct_unicode_to_str(item, ignore_dicts=True)
                for item in data
            ]

        # if this is a dictionary, return dictionary of byteified keys
        # and values but only if we haven't already byteified it
        if isinstance(data, dict) and not ignore_dicts:
            # decode datetime object separately
            data = jsondate._datetime_decoder(data)
            return {
                cls._convert_struct_unicode_to_str(key, ignore_dicts=True):
                    cls._convert_struct_unicode_to_str(value, ignore_dicts=True)
                for key, value in data.iteritems()
            }

        # if it's anything else, return it in its original form
        return data

    def get_intermediary_reference_schedule(self):
        interm_reference_file = os.path.join(self.basedir,
                                             self.intermediary_reference_file)

        with open(interm_reference_file) as fd:
            intermediary_input_dict = jsondate.load(
                fd, object_hook=self._convert_struct_unicode_to_str)

        schedule = Schedule.load_from_dict(intermediary_input_dict)
        return schedule

    def import_setup_handle_smartsheet(self):
        token = os.environ.get('SMARTSHEET_TOKEN', None)
        if not token:
            pytest.fail('You need to specify environment variable '
                        'SMARTSHEET_TOKEN to run smartsheet tests.')

        converter_options = {
            'smartsheet_token': token,
        }

        intermediary_input = self.get_intermediary_reference_schedule()

        conv = ScheduleConverter(intermediary_input)
        sheet_id = conv.export_schedule(output=None,
                                        target_format='smartsheet',
                                        options=converter_options)

        self._smartsheet_inject_extra_column(sheet_id, converter_options)

        return sheet_id, converter_options

    def _smartsheet_inject_extra_column(self, sheet_id, converter_options):
        """
        Add extra column with duplicated name. For testing only.

        Value of this extra column are shifted by 3 days into future from
        original.
        """
        client = Smartsheet(converter_options['smartsheet_token'])
        sheet = client.Sheets.get_sheet(sheet_id, page_size=None, page=None)

        start_column_id = None

        for column in sheet.columns:
            if column.title in ['Start', 'Start Date']:
                start_column_id = column.id
                break

        # Add intentionally another Start Date column - just for test purpose
        column_startdate_dup = client.models.Column({
            'title': 'Start Date',
            'type': 'DATE',
            'index': 11,
        })
        resp = client.Sheets.add_columns(sheet_id, column_startdate_dup)

        duplicated_column_id = resp.result[0].id

        if resp.message != 'SUCCESS':
            msg = 'Adding column failed: {}'.format(resp)
            raise SmartSheetExportException(msg, source=sheet_id)

        datetime_format = '%Y-%m-%dT%H:%M:%S'
        updated_rows = []
        for row in sheet.rows:
            original_cell_value = row.get_column(start_column_id).value
            new_value = datetime.datetime.strptime(original_cell_value,
                                                   datetime_format)
            new_value += datetime.timedelta(days=3)

            cell = client.models.Cell()
            cell.column_id = duplicated_column_id
            cell.value = new_value.strftime(datetime_format)

            new_row = client.models.Row()
            new_row.id = row.id
            new_row.cells.append(cell)
            updated_rows.append(new_row)

        resp = client.Sheets.update_rows(sheet_id, updated_rows)
        if resp.message != 'SUCCESS':
            msg = 'Inserting duplicated cells failed: {}'.format(resp)
            raise SmartSheetExportException(msg, source=sheet_id)

    def import_teardown_handle_smartsheet(self, handle, converter_options):
        client = Smartsheet(converter_options['smartsheet_token'])
        client.Sheets.delete_sheet(handle)

    def import_assert_changelog_smartsheet(self,
                                           reference_schedule_dict,
                                           imported_schedule_dict):
        changelog = imported_schedule_dict['changelog']
        assert len(changelog.keys()) == 1
        assert isinstance(changelog.keys()[0], int)

        record = changelog.values()[0]
        date_now = datetime.datetime.now(tz=tzutc())
        assert self.test_import_start_timestamp < record['date']
        assert record['date'] < date_now

    def test_import(self, handler_name, import_schedule_file):
        """
        Generic pytest fixture that provides arguments for import_schedule meth.
        If there is specified 'import_schedule_file' value of argument,
        its considered as filename of file-based handle.
        Otherwise will try to call 'fixture_import_handlename' method that
        provides these arguments (handle, options).
        Also it works as setup/teardown-like method of the handle's test
        (using yield within the 'fixture_import_' method).
        """
        handle = None
        converter_options = dict()
        self.test_import_start_timestamp = datetime.datetime.now(tz=tzutc())

        if import_schedule_file:
            handle = os.path.join(self.basedir, self.schedule_files_dir,
                                  import_schedule_file)
            converter_options = {
                'source_storage_format': 'local'
            }
        else:
            callback_name = 'import_setup_handle_' + handler_name
            if hasattr(self, callback_name):
                import_setup_fn = getattr(self, callback_name)
                handle, converter_options = import_setup_fn()

        try:
            conv = ScheduleConverter()
            schedule = conv.import_schedule(handle,
                                            schedule_src_format=handler_name,
                                            options=converter_options)
            assert 0 == len(schedule.errors_import)

            imported_schedule_dict = schedule.dump_as_dict()
            self._clean_interm_struct(imported_schedule_dict)

            interm_reference_file = os.path.join(
                self.basedir, self.intermediary_reference_file)
            regenerate = os.environ.get('REGENERATE', False) == 'true'
            if regenerate:
                log.info('test_import: Regenerating interm. reference file'
                         'from imported schedule.')

                with open(interm_reference_file, 'w+') as fd:
                    jsondate.dump(imported_schedule_dict,
                                  fd,
                                  sort_keys=True,
                                  indent=4,
                                  separators=(',', ': '))

            with open(interm_reference_file) as fd:
                reference_dict = json.load(
                    fd, object_hook=self._convert_struct_unicode_to_str)
            self._clean_interm_struct(reference_dict)

            # test/assert changelog separately, if there exists a hook
            callback_name = 'import_assert_changelog_' + handler_name
            if hasattr(self, callback_name):
                assert_changelog_fn = getattr(self, callback_name)
                assert_changelog_fn(reference_dict, imported_schedule_dict)

                # If asserting of changelog went well,
                # drop it from schedules dicts
                imported_schedule_dict.pop('changelog', None)
                reference_dict.pop('changelog', None)

            assert reference_dict == imported_schedule_dict
            
        finally:
            callback_name = 'import_teardown_handle_' + handler_name
            if not import_schedule_file and hasattr(self, callback_name):
                import_setup_fn = getattr(self, callback_name)
                import_setup_fn(handle, converter_options)

    def test_export(self, handler_name, export_schedule_file, 
                    flat, flag_show, flag_hide, options,
                    tmpdir):
        full_export_schedule_file = os.path.join(self.basedir,
                                                 self.schedule_files_dir,
                                                 export_schedule_file)
        intermediary_input = self.get_intermediary_reference_schedule()
        export_output_file = tmpdir.join('exported_file')
        export_output_filename = str(export_output_file)

        conv = ScheduleConverter(intermediary_input)
        
        if flat:
            conv.schedule.make_flat()
        
        conv.schedule.filter_flags(flag_show, flag_hide)
        
        conv.export_schedule(export_output_filename, handler_name,
                             options=options)
        actual_output = export_output_file.read()

        regenerate = os.environ.get('REGENERATE', False) == 'true'

        if regenerate:
            log.info('test_export: Regenerating exported file from '
                     'reference schedule.')
            shutil.copy(export_output_filename, full_export_schedule_file)

        with open(full_export_schedule_file) as fd:
            expected_output = fd.read()
        
        # sanitize if needed
        sanitize_fn_name = '_sanitize_export_test_{}'.format(handler_name)
        if hasattr(self, sanitize_fn_name):
            sanitize_func = getattr(self, sanitize_fn_name)
            expected_output = sanitize_func(expected_output)
            actual_output = sanitize_func(actual_output)

        assert expected_output == actual_output




import logging
import pytest

from lxml import etree

from schedules_tools.schedule_handlers import msp
from schedules_tools import models

logging.basicConfig()


class TestUnit_msp_parseExtAttrs(object):
    schedule = None
    task = None

    def _prepare_inject_value(self):
        self.schedule = models.Schedule()
        self.task = models.Task(schedule=self.schedule)

    def _inject_value(self, value):
        self.task.parse_extended_attr(value)

    def test_link(self):
        url = 'http://www.somewhere.io'
        valid_patterns = [
            'link: {}',
            'Link: {}',
            'Link:{}',
            'Link :{}',
            'link:{}',
            'Link:         {}',
            'Link:\n {}',
            '   Link:\t {} ',
            '   link\n:\t {} ',
        ]

        invalid_patterns = [
            '{}',       # missing prefix 'Link: '
            'link {}',  # missing ':'
        ]

        for pattern in valid_patterns:
            self._prepare_inject_value()
            self._inject_value(pattern.format(url))
            assert self.task.link == url

        for pattern in invalid_patterns:
            self._prepare_inject_value()
            self._inject_value(pattern.format(url))
            assert self.task.link != url

    def test_note(self):
        note = 'Very important text'
        valid_patterns = [
            'note: {}',
            'Note: {}',
            'Note:{}',
            'Note :{}',
            'note:{}',
            'Note:         {}',
            'Note:\n {}',
            '   Note:\t {} ',
            '   note\n:\t {} ',
        ]

        invalid_patterns = [
            '{}',
        ]

        for pattern in valid_patterns:
            self._prepare_inject_value()
            self._inject_value(pattern.format(note))
            assert self.task.note == note

        for pattern in invalid_patterns:
            self._prepare_inject_value()
            self._inject_value(pattern.format(note))
            assert self.task.note != note

    def test_flags(self):
        flags = set(['qe', 'pm', 'dev'])

        valid_patterns = [
            'flags: qe, pm, dev',
            'Flags: qe pm dev',
            'Flags:dev , pm , qe',
            'FLAgs : qe, pm dev',
            'flags:QE, PM, DEV',
            'Flags:\n qe, pm, dev',
            '   FLags:\t qe,\n pm, \tdev ',
            '  flags\n:\t dev PM qe, ',
        ]

        invalid_patterns = [
            'flag: qe pm dev',      # missing 's' in 'flags'
            'flags qe, pm, devel',  # missing ':'
        ]

        invalid_patterns_schedule = [
            'flags: qe; pm; dev',   # incorrect separator ';'
            'flags: qe, pm, devel'  # not expected value 'devel'
        ]

        for pattern in valid_patterns:
            self._prepare_inject_value()
            assert self.schedule.used_flags == set()
            self._inject_value(pattern)
            assert set(self.task.flags) == flags
            assert self.schedule.used_flags == flags

        for pattern in invalid_patterns:
            self._prepare_inject_value()
            assert self.schedule.used_flags == set()
            self._inject_value(pattern)
            assert set(self.task.flags) != flags
            assert self.schedule.used_flags == set()

        # flags will be parsed correctly, but their content wil not match
        # to reference
        for pattern in invalid_patterns_schedule:
            self._prepare_inject_value()
            assert self.schedule.used_flags == set()
            self._inject_value(pattern)
            assert set(self.task.flags) != flags
            assert self.schedule.used_flags != set()



import datetime
import pytest
import os
import jsondate

from schedules_tools.tests import create_test_schedule
from schedules_tools.converter import ScheduleConverter
from schedules_tools.diff_v2 import ScheduleDiff

BASE_DIR = os.path.dirname(os.path.realpath(__file__))


class TestDiff(object):
    sch1 = None
    sch2 = None

    @pytest.fixture(autouse=True)
    def setUp(self):
        self.sch1 = create_test_schedule()
        self.sch2 = create_test_schedule()

    def test_identical(self):
        diff = self.sch1.diff(self.sch2)

        assert diff == ''

    def test_sch_attrs(self):
        self.sch1.name = '_test_value 1.0.42'
        self.sch1.used_flags = set(['flag1', 'qe'])
        self.sch1.dStart = datetime.datetime(1970, 1, 1)
        self.sch1.dFinish = datetime.datetime(4970, 1, 1)
        self.sch2.changelog['today'] = 'change'

        diff = self.sch1.diff(self.sch2)

        assert diff != ''
        assert 'name' in diff
        assert 'used_flags' in diff
        assert 'dStart' in diff
        assert 'dFinish' in diff
        assert 'changelog' in diff

    def test_task_attrs(self):
        date_delta = datetime.timedelta(1)

        task = self.sch1.tasks[0].tasks[0]
        task.dStart -= date_delta
        task.dFinish += date_delta
        task.link = 'nothing'
        task.note = 'note123'

        diff = self.sch1.diff(self.sch2)

        assert 'link' in diff
        assert 'dStart' in diff
        assert 'dFinish' in diff
        assert 'note' in diff

    def test_missing_task(self):
        diff = self.sch1.diff(self.sch2)
        assert diff == ''

        self.sch2.tasks[0].tasks.pop()

        diff = self.sch1.diff(self.sch2)
        assert diff != ''

    def test_whole_days(self):
        time_delta = datetime.timedelta(minutes=1)
        self.sch2.tasks[0].tasks[1].tasks[0].dFinish -= time_delta

        diff = self.sch1.diff(self.sch2, whole_days=True)
        assert diff == ''


def scheds_list(scheds_dir, ref_sched):
    return [f for f in os.listdir(scheds_dir) if f != ref_sched]


class TestScheduleDiff(object):
    ref_sched = 'sched_diff_reference.xml'

    SCHEDS_DIR = os.path.join(BASE_DIR, 'schedule_files/schedule_diff')
    OUTPUT_FILES_DIR = os.path.join(BASE_DIR, 'fixtures/schedule_diff')

    def import_schedule(self, filename):
        path = os.path.join(self.SCHEDS_DIR, filename)
        conv = ScheduleConverter()
        return conv.import_schedule(path)

    @pytest.fixture(params=scheds_list(scheds_dir=SCHEDS_DIR, 
                                       ref_sched=ref_sched), scope='class')
    def diff_res(self, request):
        filename = request.param

        left = self.import_schedule(self.ref_sched)
        right = self.import_schedule(filename)

        return ScheduleDiff(left, right), filename

    @pytest.fixture
    def expected(self, request, diff_res):
        name = os.path.splitext(diff_res[1])[0]
        ext = request.param

        with open(os.path.join(self.OUTPUT_FILES_DIR, '.'.join([name, ext]))) as f:
            yield f

    @pytest.mark.parametrize('expected', ['txt'], indirect=True)
    def test_txt_output(self, diff_res, expected):
        assert str(diff_res[0]) == expected.read()

    @pytest.mark.parametrize('expected', ['json'], indirect=True)
    def test_json_output(self, diff_res, expected):
        diff_json = diff_res[0].dump_json()
        assert jsondate.loads(diff_json) == jsondate.load(expected)


class TestDiffCLI(object):
    pass



#!/usr/bin/python
# -*- coding: utf-8 -*-
import os

from setuptools import setup, find_packages
from shutil import copyfile

from scripts.setup_utils import read_version, write_version, get_rpm_version
from scripts.rpm_log import get_rpm_log


project_name = "schedules-tools"
save_version_dirs = ["schedules_tools"]
project_url = "https://github.com/RedHat-Eng-PGM/python-schedules-tools"
project_author = "Red Hat, Inc."
project_author_email = "pp-dev-list@redhat.com"
project_description = "Schedules tools to handle/convert various schedule formats"
package_name = project_name
package_version = ['5.30.0', 1, 'git']  # default


# VERSION - write version only when building from git
if os.path.isdir(".git"):
    # we're building from a git repo -> store version tuple to __init__.py
    if package_version[2] == "git":
        force = True
        package_version[0], package_version[1], package_version[2] = get_rpm_version()

    for i in save_version_dirs:
        file_name = os.path.join(i, "version.py")
        write_version(file_name, package_version)
else:  # see if there is already saved version
    for i in save_version_dirs:
        file_name = os.path.join(i, "version.py")
        version = read_version(file_name)
        if version:
            package_version = list(version)
            break
    

# prepare specfile if existing
spec = 'python-{}.spec'.format(project_name)
spec_in = 'spec/{}.in'.format(spec)
spec_tgt = 'dist/' + spec
if os.path.exists(spec_in):
    if not os.path.exists('dist'):
        os.mkdir('dist')
        
    copyfile(spec_in, spec_tgt)
    
    with open(spec_tgt, 'r+') as f:
        orig_content = f.read()
        f.seek(0, 0)
        f.write('%define version {}\n%define release_number {}\n%define checkout {}\n'.format(
            *package_version))
        f.write(orig_content)
        f.write(get_rpm_log())

with open('requirements.txt') as fd:
    requirements = fd.read().split()


setup(
    name=package_name,
    url=project_url,
    version='.'.join([str(v) for v in package_version[0:2]]),
    license='GPL',
    author=project_author,
    author_email=project_author_email,
    description=project_description,
    packages=find_packages(exclude=('scripts',)),
    include_package_data=True,
    test_suite='tests',
    tests_require=['testtools'],
    install_requires=requirements,
    scripts=[
        'scripts/schedule-convert',
        'scripts/schedule-diff'],
)



import logging
import os

from schedules_tools import discovery
from schedules_tools import SchedulesToolsException
from schedules_tools.discovery import search_paths
from schedules_tools.models import Schedule
from schedules_tools.storage_handlers import AcquireLockException


log = logging.getLogger(__name__)


class ScheduleFormatNotSupported(Exception):
    pass


class StorageFormatNotSupported(Exception):
    pass


class HandleWithoutExport(Exception):
    pass


class HandlerMissingDeps(Exception):
    pass


class ScheduleConverter(object):
    """
    Abstraction class to work with handles/schedules
    no matter the exact handler/schedule type.
    """
    schedule = None
    storage_handler = None
    local_handle = None

    def __init__(self, schedule=None):
        self.schedule = schedule

    # TODO: take a look if _get_handler* methods can't be shared for both
    # schedule/storage
    @staticmethod
    def _get_schedule_handler_for_handle(handle):
        for module in discovery.schedule_handlers.values():
            if module['class'].handle_deps_satisfied and module['class'].is_valid_source(handle):
                return module

        msg = "Can't find schedule handler for handle: {}".format(handle)
        raise ScheduleFormatNotSupported(msg)

    @staticmethod
    def _get_schedule_handler_for_format(format):
        if format not in discovery.schedule_handlers.keys():
            msg = "Can't find schedule handler for format: {}".format(format)
            raise ScheduleFormatNotSupported(msg)

        return discovery.schedule_handlers[format]

    @staticmethod
    def _get_storage_handler_for_format(format):
        if format not in discovery.storage_handlers.keys():
            msg = "Can't find storage handler for format: {}".format(format)
            raise StorageFormatNotSupported(msg)

        return discovery.storage_handlers[format]

    @classmethod
    def _get_schedule_handler_struct(cls, handle=None, format=None):
        if format:
            handler_struct = cls._get_schedule_handler_for_format(format)
        else:
            handler_struct = cls._get_schedule_handler_for_handle(handle)

        return handler_struct

    @classmethod
    def _get_schedule_handler_cls(cls, *args, **kwargs):
        return cls._get_schedule_handler_struct(*args, **kwargs)['class']

    @classmethod
    def _get_storage_handler_cls(cls, *args, **kwargs):
        return cls._get_storage_handler_for_format(*args, **kwargs)['class']

    def _init_storage_handler(self, handle, storage_src_format, options=dict()):
        """Prepare storage handler if it's necessary and isn't already prepared"""
        if storage_src_format and not self.storage_handler:
            storage_handler_cls = self._get_storage_handler_cls(storage_src_format)
            storage_handler = storage_handler_cls(handle=handle,
                                                  options=options)
            self.storage_handler = storage_handler
        
        return self.storage_handler

    def _get_local_handle_from_storage(self, handle, storage_src_format, options):
        """Init storage and get local handle from storage if specified otherwise return original handle"""
        if not self.local_handle:
            self.local_handle = handle
    
            self._init_storage_handler(handle, storage_src_format, options)
            
            if self.storage_handler:
                self.local_handle = self.storage_handler.get_local_handle()

        return self.local_handle

    def cleanup_local_handle(self):
        if self.storage_handler:
            self.storage_handler.clean_local_handle()
        
        self.local_handle = None

    # Following methods call their counterparts on handlers
    def handle_modified_since(self, 
                              handle, 
                              mtime,
                              schedule_src_format=None, 
                              storage_src_format=None,
                              cleanup=True,
                              options=dict()
                              ):
        """ Return boolean (call schedule_handler specific method) to be able to bypass processing """
        errors = []
        handle_modified = True

        self._init_storage_handler(handle, storage_src_format, options)

        # use storage if possible, otherwise schedule
        if self.storage_handler and self.storage_handler.provide_mtime:
            handle_modified = self.storage_handler.handle_modified_since(mtime)
        else:
            local_handle = self._get_local_handle_from_storage(handle, 
                                                               storage_src_format,
                                                               options
                                                               )        

            schedule_handler_cls = self._get_schedule_handler_cls(
                                            handle=local_handle, 
                                            format=schedule_src_format
                                            )

            if not schedule_handler_cls.handle_deps_satisfied:
                msg = ('Schedule handler {} has unsatisfied dependencies, '
                       'can\'t get modified time.'.format(schedule_handler_cls))
                raise HandlerMissingDeps(msg)
    
            try:
                handle_modified = schedule_handler_cls(handle=local_handle, 
                                                       options=options
                                                       ).handle_modified_since(mtime)
            except SchedulesToolsException as e:
                error_item = e.__class__.__name__, str(e).split('\n'), e.source
                errors.append(error_item)
            
            if cleanup:
                self.cleanup_local_handle()
            
        return handle_modified, errors

    def import_schedule(self,
                        handle, 
                        schedule_src_format=None,
                        storage_src_format=None,
                        cleanup=True,
                        options=dict()
                        ):
        
        # convert to local handle if needed
        local_handle = self._get_local_handle_from_storage(handle, 
                                                           storage_src_format,
                                                           options
                                                           )
        schedule_handler_cls = self._get_schedule_handler_cls(
                                        handle=local_handle, 
                                        format=schedule_src_format
                                        )
        if not schedule_handler_cls.handle_deps_satisfied:
            msg = ('Schedule handler {} has unsatisfied dependencies, '
                   'can\'t import schedule.'.format(schedule_handler_cls))
            raise HandlerMissingDeps(msg)

        schedule_handler = schedule_handler_cls(
                                        handle=local_handle, 
                                        options=options
                                        )

        # imports changelog and mtime - if implemented
        schedule = Schedule()
        try:
            schedule = schedule_handler.import_schedule()

            # if storage defined and provides changelog/mtime - use storage handler to overwrite it
            if self.storage_handler:
                if self.storage_handler.provide_changelog:
                    schedule.changelog = self.storage_handler.get_handle_changelog()

                if self.storage_handler.provide_mtime:
                    schedule.mtime = self.storage_handler.get_handle_mtime()

        except SchedulesToolsException as e:
            error_item = e.__class__.__name__, str(e).split('\n'), e.source
            schedule.errors_import.append(error_item)

        self.schedule = schedule
        
        if cleanup:
            self.cleanup_local_handle()
        
        return self.schedule

    def export_schedule(self, output, target_format, update_filename=False, options=dict()):
        schedule_slug = options.get('slug', '')

        schedule_handler_cls = self._get_schedule_handler_cls(format=target_format)

        if not schedule_handler_cls.provide_export:
            raise HandleWithoutExport(
                'Schedule handler for {} doesn\'t provide export.'
                .format(target_format))

        if not schedule_handler_cls.handle_deps_satisfied:
            msg = ('Schedule handler {} has unsatisfied dependencies, '
                   'can\'t export schedule.'.format(schedule_handler_cls))
            raise HandlerMissingDeps(msg)

        schedule_handler = schedule_handler_cls(schedule=self.schedule, options=options)
        
        if update_filename and schedule_handler.default_export_ext:  # change/add export extension according to handler
            output = '.'.join([os.path.splitext(output)[0], 
                               schedule_handler.default_export_ext])

        return schedule_handler.export_schedule(output)
        
        

def convert(args):
    opt_args = vars(args)
    
    # set timezone - it's desirable that all calculations are made in same TZ
    os.environ['TZ'] = opt_args['tz']
    
    for path in opt_args.pop('handlers_path'):
        search_paths.append(path)

    converter = ScheduleConverter()

    try:
        converter.import_schedule(handle=args.source,
                                  schedule_src_format=args.source_format,
                                  storage_src_format=args.source_storage_format,
                                  options=opt_args)

        if converter.schedule.errors_import:
            for err in converter.schedule.errors_import:
                msg = '{} Handle: {}\n{}'.format(err[0], err[2], '\n'.join(err[1]))
                log.error(msg)
            return False
        
    except AcquireLockException as e:
        log.error(e)
        return

    check_tasks = dict()
    for task_name in args.check_taskname:
        check_tasks[task_name] = False

    for task_name in args.check_taskname_startswith:
        check_tasks[task_name] = True

    if args.check_taskname or args.check_taskname_startswith:
        missing_tasks = converter.schedule.check_for_taskname(check_tasks)
        if missing_tasks:
            log.info('Missing tasks: {}'.format(list(missing_tasks)))

    if args.flat or args.milestones:
        converter.schedule.make_flat()
        
    if args.milestones:
        converter.schedule.filter_milestones()
        
    flag_show = args.flag_show.split(',')
    if flag_show == ['']:
        flag_show = []        

    flag_hide = [f for f in args.flag_hide.split(',') if f]
    if flag_hide == ['']:
        flag_hide = []  
        
    converter.schedule.filter_flags(flag_show, flag_hide)      


    # do we have target name defined?
    update_filename = False
    if not args.target:
        args.target = args.source
        update_filename = True

    converter.export_schedule(args.target,
                              args.target_format,
                              update_filename=update_filename,
                              options=opt_args)   


def get_handlers_args_parser(add_help=False):
    '''Return parent parser for schedules tools scripts with handler arguments'''
    import argparse 
    
    parser = argparse.ArgumentParser(add_help=add_help)
    
    parser.add_argument('--log-level',
                        help='ERROR (default) | WARN | INFO | DEBUG',
                        default='ERROR')

    parser.add_argument('--handlers-path', 
                        help='Add python-dot-notation path to discover handlers (needs to '
                             'be python module), can be called several times '
                             '(conflicting names will be overriden - the last '
                             'implementation will be used)',
                        action='append',
                        default=[],
                        )

    parser.add_argument('-f', '--force',
                        help='Force target overwrite',
                        default=False,
                        action='store_true')

    parser.add_argument('--slug', metavar='SCHEDULE_SLUG',
                        help='Override schedule slug (e.g. rhel)')
    parser.add_argument('--use-tji-file',
                        help='Use TJI file when exporting into TJP',
                        default=False,
                        action='store_true')
    parser.add_argument('--tjp-keep-tjx',
                        help='Keep tjx intermediate file when importing from TJP',
                        default=False,
                        action='store_true')

    parser.add_argument('--rally-iter', help='Rally iteration to import',
                        default='')

    parser.add_argument('--check-taskname',
                        help='Check existence given task name as exact match '
                             '(can be used multiple times)',
                        action='append', metavar='TASKNAME', default=[])
    parser.add_argument('--check-taskname-startswith',
                        help='Check existence given task by matching beginning '
                             'name (can be used multiple times)',
                        action='append', metavar='TASKNAME_STARTSWITH',
                        default=[])

    parser.add_argument('--source-storage-format',
                        choices=discovery.storage_handlers.keys(),
                        metavar='SRC_STORAGE_FORMAT',
                        help='Source storage format to use')
    
    parser.add_argument('--cvs-repo-name',
                        help='Name of CVS repository to checkout')
    parser.add_argument('--cvs-root',
                        help='Root of CVS repository')
    parser.add_argument('--cvs-checkout-path',
                        help='Path to shared working copy of CVS repository')
    parser.add_argument('--cvs-exclusive-access',
                        help='Restrict to run just one CVS command at the same time',
                        action='store_true')
    parser.add_argument('--cvs-lock-redis-uri',
                        help='Redis URI that is required by --cvs-exclusive-access, default: localhost:6379/0',
                        default='localhost:6379/0')
    
    parser.add_argument('--smartsheet-token',
                        help='Access token for using SmartSheet API')
    
    parser.add_argument('--date-format',
                        help='Date format used for export where applicable')    

    parser.add_argument('--html-title',
                        help='HTML export page title')    
    parser.add_argument('--html-table-header',
                        help='HTML export table header')    

    
    parser.add_argument('--flat',
                        help='Make output schedule flat',
                        default=False,
                        action='store_true')

    parser.add_argument('--milestones',
                        help='Filter only milestones (implies --flat)',
                        default=False,
                        action='store_true')
    
    parser.add_argument('--flag-show',
                        help='Filter schedule - show tasks with any of these flags',
                        default='',
                        )
    parser.add_argument('--flag-hide',
                        help='Filter schedule - hide tasks with any of these flags, hide has preferrence over show',
                        default='',
                        )


    parser.add_argument('--source-format',
                        metavar='SRC_FORMAT',
                        help='Source format to enforce')

    parser.add_argument('--target-format',
                        metavar='TARGET_FORMAT',
                        help='Target format to convert',
                        default='html')
    
    parser.add_argument('--tz',
                        help='Timezone used for schedule conversions (default "America/New_York")',
                        default='America/New_York')
    
    return parser
    
